__all__ = ['CheckMultisigMethod',
           'CheckSigMethod',
           'Hash160Method',
           'Hash256Method',
           'Murmur32Method',
           'NamedCurveType',
           'Ripemd160Method',
           'Sha256Method',
           'VerifyWithECDsaMethod',
           ]

from boa3.model.builtin.interop.crypto.checkmultisigmethod import CheckMultisigMethod
from boa3.model.builtin.interop.crypto.checksigmethod import CheckSigMethod
from boa3.model.builtin.interop.crypto.hash160method import Hash160Method
from boa3.model.builtin.interop.crypto.hash256method import Hash256Method
from boa3.model.builtin.interop.crypto.murmur32method import Murmur32Method
from boa3.model.builtin.interop.crypto.namedcurvetype import NamedCurveType
from boa3.model.builtin.interop.crypto.ripemd160method import Ripemd160Method
from boa3.model.builtin.interop.crypto.sha256method import Sha256Method
from boa3.model.builtin.interop.crypto.verifywithecdsa import VerifyWithECDsaMethod



from typing import Dict

from boa3.model.builtin.interop.iterator.iteratortype import IteratorType
from boa3.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.model.variable import Variable


class GetAllCandidatesMethod(NeoContractMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'get_all_candidates'
        native_identifier = 'getAllCandidates'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, native_identifier, args,
                         return_type=IteratorType.build(Type.tuple.build_collection(
                             [ECPointType.build(), Type.int]
                         )))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

from boa3.model.operation.operation import IOperation
from boa3.model.type.itype import IType


class BinaryOperation(IOperation, ABC):
    """
    An interface used to represent binary operations

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = []

    def __init__(self, left: IType, right: IType = None):
        if right is None:
            right = left

        self.left_type: IType = left
        self.right_type: IType = right
        result = self._get_result(left, right)
        super().__init__(self.operator, result)

    @property
    def number_of_operands(self) -> int:
        return 2

    @property
    def is_symmetric(self) -> bool:
        return False

    @abstractmethod
    def _get_result(self, left: IType, right: IType) -> IType:
        """
        Gets the result type of the operation given the operands types.

        :param left: left operand type
        :param right: right operand type
        :return: the result type of the operation. Type.none if the operands are not valid.
        """
        pass

    @classmethod
    def build(cls, *operands: IType) -> Optional[BinaryOperation]:
        if len(operands) == 1:
            return cls._build_with_left_arg(operands[0])
        if len(operands) == 2:
            return cls._build_with_two_args(operands[0], operands[1])

    @classmethod
    def _build_with_left_arg(cls, left: IType) -> Optional[BinaryOperation]:
        """
        Creates a binary operation with the given operands types

        :param left: left operand type
        :return: The built operation if the operands is valid. None otherwise
        :rtype: BinaryOperation or None
        """
        return cls(left)

    @classmethod
    def _build_with_two_args(cls, left: IType, right: IType) -> Optional[BinaryOperation]:
        """
        Creates a binary operation with the given operands types

        :param left: left operand type
        :param right: right operand type
        :return: The built operation if the operands are valid. None otherwise
        :rtype: BinaryOperation or None
        """
        operation = cls(left, right)
        if operation.validate_type(left, right):
            return operation
        return cls(left, left)

    def get_valid_operand_for_validation(self, left_operand: IType,
                                         right_operand: IType = None) -> Tuple[Optional[IType], Optional[IType]]:
        """
        Returns a valid operand type for the given types.

        :param left_operand: reference type
        :param right_operand: reference type
        """
        return next((valid_type for valid_type in self._valid_types if valid_type.is_type_of(left_operand)),
                    None), self.right_type



__all__ = ['Addition',
           'Concat',
           'Division',
           'FloorDivision',
           'ListAddition',
           'Modulo',
           'Multiplication',
           'Power',
           'StrBytesMultiplication',
           'Subtraction'
           ]

from boa3.model.operation.binary.arithmetic.addition import Addition
from boa3.model.operation.binary.arithmetic.concat import Concat
from boa3.model.operation.binary.arithmetic.division import Division
from boa3.model.operation.binary.arithmetic.floordivision import FloorDivision
from boa3.model.operation.binary.arithmetic.listaddition import ListAddition
from boa3.model.operation.binary.arithmetic.modulo import Modulo
from boa3.model.operation.binary.arithmetic.multiplication import Multiplication
from boa3.model.operation.binary.arithmetic.power import Power
from boa3.model.operation.binary.arithmetic.strbytesmultiplication import StrBytesMultiplication
from boa3.model.operation.binary.arithmetic.subtraction import Subtraction



import os
from typing import Any, Dict, Iterable, Optional, Tuple, Type, Union
from unittest import TestCase

from boa3 import constants, env
from boa3.analyser.analyser import Analyser
from boa3.compiler.compiler import Compiler
from boa3.exception.CompilerError import CompilerError
from boa3.model.method import Method
from boa3.neo.smart_contract.VoidType import VoidType
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3.neo3.vm import VMState
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine
from boa3_test.tests.test_classes.transactionattribute import oracleresponse


class BoaTest(TestCase):
    dirname: str = None

    ABORTED_CONTRACT_MSG = 'ABORT is executed'
    ARGUMENT_OUT_OF_RANGE_MSG_PREFIX = 'Specified argument was out of the range of valid values.'
    ASSERT_RESULTED_FALSE_MSG = 'ASSERT is executed with false result.'
    CANT_CALL_SYSCALL_WITH_FLAG_MSG_PREFIX = "Cannot call this SYSCALL with the flag"
    CANT_FIND_METHOD_MSG_PREFIX = "Can't find method"
    CANT_PARSE_VALUE_MSG = "The value could not be parsed."
    CALLED_CONTRACT_DOES_NOT_EXIST_MSG = 'Called Contract Does Not Exist'
    GAS_MUST_BE_POSITIVE_MSG = 'GAS must be positive.'
    GIVEN_KEY_NOT_PRESENT_IN_DICT_MSG_REGEX = "The given key '\\S+' was not present in the dictionary."
    MAP_KEY_NOT_FOUND_ERROR_MSG = 'Key not found in Map'
    MAX_ITEM_SIZE_EXCEED_MSG_PREFIX = 'MaxItemSize exceed'
    METHOD_DOESNT_EXIST_IN_CONTRACT_MSG_REGEX_PREFIX = r'^Method "\S+" with \d+ parameter\(s\) ' \
                                                       r"doesn't exist in the contract"
    NULL_POINTER_MSG = 'Object reference not set to an instance of an object.'
    UNHANDLED_EXCEPTION_MSG_PREFIX = "An unhandled exception was thrown."
    VALUE_CANNOT_BE_NEGATIVE_MSG = 'value can not be negative'
    VALUE_DOES_NOT_FALL_WITHIN_EXPECTED_RANGE_MSG = 'Value does not fall within the expected range.'
    VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX = r'The value( \S*)? is out of range.$'

    default_folder: str = ''

    @classmethod
    def setUpClass(cls):
        folders = os.path.abspath(__file__).split(os.sep)
        cls.dirname = '/'.join(folders[:-3])
        cls.test_root_dir = '/'.join(folders[-3:-2])
        cls.default_test_folder = ('{0}/{1}'.format(cls.test_root_dir, cls.default_folder)
                                   if len(cls.default_folder) else cls.test_root_dir)

        super(BoaTest, cls).setUpClass()

    def get_compiler_analyser(self, compiler: Compiler) -> Analyser:
        return compiler._analyser

    def get_all_imported_methods(self, compiler: Compiler) -> Dict[str, Method]:
        from boa3.compiler.filegenerator import FileGenerator
        generator = FileGenerator(compiler.bytecode, compiler._analyser, compiler._entry_smart_contract)
        return {constants.VARIABLE_NAME_SEPARATOR.join(name): value for name, value in generator._methods_with_imports.items()}

    def indent_text(self, text: str, no_spaces: int = 4) -> str:
        import re
        return re.sub('\n[ \t]+', '\n' + ' ' * no_spaces, text)

    def assertCompilerLogs(self, expected_logged_exception, path) -> Union[bytes, str]:
        output, error_msg = self.assertCompilerLogsError(expected_logged_exception, path)
        if not issubclass(expected_logged_exception, CompilerError):
            return output
        else:
            # filter to get only the error message, without location information
            import re

            result = re.search('^\\d+:\\d+ - (?P<msg>.*?)\t\\W+\\<.*?\\>', error_msg)
            try:
                return result.group('msg')
            except BaseException:
                return output

    def assertCompilerLogsError(self, expected_logged_exception, path):
        output = None
        with self.assertLogs() as log:
            from boa3.exception.NotLoadedException import NotLoadedException
            try:
                from boa3.boa3 import Boa3
                output = Boa3.compile(path)
            except NotLoadedException:
                # when an compiler error is logged this exception is raised.
                pass

        for logger in log.records:
            import logging
            logging.log(level=logger.levelno, msg=logger.msg)

        expected_logged = [exception for exception in log.records
                           if isinstance(exception.msg, expected_logged_exception)]
        if len(expected_logged) < 1:
            raise AssertionError('{0} not logged'.format(expected_logged_exception.__name__))
        return output, expected_logged[0].message

    def assertIsVoid(self, obj: Any):
        if obj is not VoidType:
            self.fail('{0} is not Void'.format(obj))

    def get_dir_path(self, *args: str) -> str:
        type_error_message = 'get_contract_path() takes {0} positional argument but {1} were given'
        num_args = len(args)
        if num_args == 0:
            raise TypeError(type_error_message.format(2, num_args + 1))
        if num_args > 2:
            raise TypeError(type_error_message.format(3, num_args + 1))

        values = [None, env.PROJECT_ROOT_DIRECTORY]
        for index, value in enumerate(reversed(args)):
            values[index] = value

        dir_folder, root_path = values
        return '{0}/{1}'.format(root_path, dir_folder)

    def get_contract_path(self, *args: str) -> str:
        """
        Usages:
            get_contract_path(contract_name)
            get_contract_path(dir_folder, contract_name)
            get_contract_path(root_path, dir_folder, contract_name)
        """
        type_error_message = 'get_contract_path() takes {0} positional argument but {1} were given'
        num_args = len(args)
        if num_args == 0:
            raise TypeError(type_error_message.format(2, num_args + 1))
        if num_args > 3:
            raise TypeError(type_error_message.format(4, num_args + 1))

        values = [None, self.default_test_folder, env.PROJECT_ROOT_DIRECTORY]
        for index, value in enumerate(reversed(args)):
            values[index] = value

        contract_name, dir_folder, root_path = values

        from os import path
        if not path.exists(dir_folder) and root_path is env.PROJECT_ROOT_DIRECTORY:
            path_folder = '{0}/{1}'.format(root_path, dir_folder)
            if not path.exists(path_folder) and not dir_folder.startswith(self.test_root_dir):
                path_folder = '{0}/{1}/{2}'.format(root_path, self.test_root_dir, dir_folder)
                if not path.exists(path_folder):
                    path_folder = '{0}/{1}/{2}'.format(root_path, self.default_test_folder, dir_folder)

            dir_folder = path_folder
        else:
            if path.exists(dir_folder):
                dir_folder = os.path.abspath(dir_folder)
            else:
                dir_folder = '{0}/{1}'.format(root_path, dir_folder)

        if not contract_name.endswith('.py'):
            contract_name = contract_name + '.py'

        path = '{0}/{1}'.format(dir_folder, contract_name)
        if not os.path.isfile(path):
            raise FileNotFoundError(path)
        return path

    def compile_and_save(self, path: str, root_folder: str = None, debug: bool = False, log: bool = True) -> Tuple[bytes, Dict[str, Any]]:
        nef_output = path.replace('.py', '.nef')
        manifest_output = path.replace('.py', '.manifest.json')

        from boa3.boa3 import Boa3
        from boa3.neo.contracts.neffile import NefFile
        Boa3.compile_and_save(path, root_folder=root_folder, show_errors=log, debug=debug)

        with open(nef_output, mode='rb') as nef:
            file = nef.read()
            output = NefFile.deserialize(file).script

        with open(manifest_output) as manifest_output:
            import json
            manifest = json.loads(manifest_output.read())

        return output, manifest

    def get_debug_info(self, path: str) -> Optional[Dict[str, Any]]:
        debug_info_output = path.replace('.py', '.nefdbgnfo')

        if not os.path.isfile(debug_info_output):
            return None

        from zipfile import ZipFile
        with ZipFile(debug_info_output, 'r') as dbgnfo:
            import json
            debug_info = json.loads(dbgnfo.read(os.path.basename(path.replace('.py', '.debug.json'))))
        return debug_info

    def get_output(self, path: str, root_folder: str = None) -> Tuple[bytes, Dict[str, Any]]:
        nef_output = path.replace('.py', '.nef')
        if not os.path.isfile(nef_output):
            return self.compile_and_save(path, root_folder=root_folder)

        manifest_output = path.replace('.py', '.manifest.json')

        from boa3.neo.contracts.neffile import NefFile

        if not os.path.isfile(nef_output):
            output = bytes()
        else:
            with open(nef_output, mode='rb') as nef:
                file = nef.read()
                output = NefFile.deserialize(file).script

        if not os.path.isfile(manifest_output):
            manifest = {}
        else:
            with open(manifest_output) as manifest_output:
                import json
                manifest = json.loads(manifest_output.read())

        return output, manifest

    def get_bytes_output(self, path: str) -> Tuple[bytes, Dict[str, Any]]:
        nef_output = path.replace('.py', '.nef')
        if not os.path.isfile(nef_output):
            return self.compile_and_save(path)

        manifest_output = path.replace('.py', '.manifest.json')

        if not os.path.isfile(nef_output):
            output = bytes()
        else:
            with open(nef_output, mode='rb') as nef:
                output = nef.read()

        if not os.path.isfile(manifest_output):
            manifest = {}
        else:
            with open(manifest_output) as manifest_output:
                import json
                manifest = json.loads(manifest_output.read())

        return output, manifest

    def run_smart_contract(self, test_engine: TestEngine, smart_contract_path: Union[str, bytes], method: str,
                           *arguments: Any, reset_engine: bool = False,
                           fake_storage: Dict[Tuple[str, str], Any] = None,
                           signer_accounts: Iterable[bytes] = (),
                           expected_result_type: Type = None,
                           rollback_on_fault: bool = True) -> Any:

        if isinstance(smart_contract_path, str) and smart_contract_path.endswith('.py'):
            if not (os.path.isfile(smart_contract_path.replace('.py', '.nef'))
                    and os.path.isfile(smart_contract_path.replace('.py', '.manifest.json'))):
                # both .nef and .manifest.json are required to execute the smart contract
                self.compile_and_save(smart_contract_path, log=False)
            smart_contract_path = smart_contract_path.replace('.py', '.nef')
        elif isinstance(smart_contract_path, bytes):
            from boa3.neo3.core.types import UInt160
            smart_contract_path = UInt160(smart_contract_path)

        self._set_fake_data(test_engine, fake_storage, signer_accounts)
        result = test_engine.run(smart_contract_path, method, *arguments,
                                 reset_engine=reset_engine, rollback_on_fault=rollback_on_fault)

        return self._filter_result(test_engine, expected_result_type, result)

    def run_oracle_response(self, test_engine: TestEngine, request_id: int,
                            response_code: oracleresponse.OracleResponseCode,
                            oracle_result: bytes, reset_engine: bool = False,
                            fake_storage: Dict[Tuple[str, str], Any] = None,
                            signer_accounts: Iterable[bytes] = (),
                            expected_result_type: Type = None,
                            rollback_on_fault: bool = True) -> Any:

        self._set_fake_data(test_engine, fake_storage, signer_accounts)
        result = test_engine.run_oracle_response(request_id, response_code, oracle_result,
                                                 reset_engine=reset_engine, rollback_on_fault=rollback_on_fault)

        return self._filter_result(test_engine, expected_result_type, result)

    def _set_fake_data(self, test_engine: TestEngine,
                       fake_storage: Dict[Tuple[str, str], Any],
                       signer_accounts: Iterable[bytes]):

        if isinstance(fake_storage, dict):
            test_engine.set_storage(fake_storage)

        for account in signer_accounts:
            test_engine.add_signer_account(account)

    def _filter_result(self, test_engine, expected_result_type, result) -> Any:
        if test_engine.vm_state is not VMState.HALT and test_engine.error is not None:
            raise TestExecutionException(test_engine.error)

        if expected_result_type is not None:
            if expected_result_type is not str and isinstance(result, str):
                result = String(result).to_bytes()

            if expected_result_type is bool:
                if isinstance(result, bytes):
                    result = Integer.from_bytes(result, signed=True)
                if isinstance(result, int) and result in (False, True):
                    result = bool(result)

            if expected_result_type is bytearray and isinstance(result, bytes):
                result = bytearray(result)

        return result



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.primitive.ibytestringtype import IByteStringType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class LowerMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        if not isinstance(self_type, IByteStringType):
            from boa3.model.type.primitive.bytestringtype import ByteStringType
            self_type = ByteStringType.build()

        identifier = 'lower'
        args: Dict[str, Variable] = {'self': Variable(self_type)}

        super().__init__(identifier, args, return_type=self_type)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.StackItem import StackItemType
        from boa3.neo.vm.type.Integer import Integer

        upper_a = Integer(ord('A')).to_byte_array()
        upper_z = Integer(ord('Z')).to_byte_array()
        jmp_place_holder = (Opcode.JMP, b'\x01')

        initializing = [                    # initialize auxiliary values
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSH0, b''),            # index = 0
        ]

        verify_while = [                    # verifies if while is over
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            jmp_place_holder,               # jump to last statement to clean the stack if index >= len(string)
        ]

        get_substring_left = [              # gets the substring to the left of the index
            (Opcode.REVERSE3, b''),
            (Opcode.DUP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.LEFT, b''),             # substr_left = string[:index]
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.ROT, b''),
            (Opcode.ROT, b''),
        ]

        get_substring_middle = [            # gets the substring on the index
            (Opcode.OVER, b''),             # TODO: verify if string[index] < c0 when other values are implemented
            (Opcode.OVER, b''),
            (Opcode.PUSH1, b''),            # modifier = 1, since using upper is only supported with ASCII for now
            (Opcode.SUBSTR, b''),           # substr_middle = string[index:index+modifier]
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.DUP, b''),
            (Opcode.PUSHDATA1, Integer(len(upper_a)).to_byte_array() + upper_a),
            jmp_place_holder,               # jump to get the substring to the right if substr_middle value is lower than 'A'
        ]

        verify_greater_than_z = [           # verifies if substr_middle is between 'A' and 'Z'
            (Opcode.DUP, b''),
            (Opcode.PUSHDATA1, Integer(len(upper_z)).to_byte_array() + upper_z),
            jmp_place_holder,               # jump to get the substring to the right if substr_middle value is greater than 'Z'
        ]

        swap_upper_to_lower_case = [        # change middle_substr to lowercase equivalent
            (Opcode.PUSHINT8, Integer(32).to_byte_array(signed=True)),
            (Opcode.ADD, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
        ]

        jmp_to_join_substring = Opcode.get_jump_and_data(Opcode.JMPLT, get_bytes_count(verify_greater_than_z +
                                                                                       swap_upper_to_lower_case), True)
        get_substring_middle[-1] = jmp_to_join_substring

        jmp_to_join_substring = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(swap_upper_to_lower_case), True)
        verify_greater_than_z[-1] = jmp_to_join_substring

        get_substring_middle.extend(verify_greater_than_z)
        get_substring_middle.extend(swap_upper_to_lower_case)

        get_substring_right = [             # gets the substring to the right of the index
            (Opcode.ROT, b''),
            (Opcode.ROT, b''),
            (Opcode.INC, b''),
            (Opcode.NEGATE, b''),
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.RIGHT, b''),            # substr_right = string[index+modifier:]
        ]

        join_substrings = [                 # concatenate substr_left, substr_middle, substr_right
            (Opcode.CAT, b''),
            (Opcode.CAT, b''),              # substr_left + substr_middle + substr_right
            (Opcode.NIP, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.REVERSE3, b''),
            (Opcode.INC, b''),              # index ++
            # jump back to verify,
        ]

        jmp_to_verify_while = Opcode.get_jump_and_data(Opcode.JMP, -get_bytes_count(verify_while +
                                                                                    get_substring_left +
                                                                                    get_substring_middle +
                                                                                    get_substring_right +
                                                                                    join_substrings))
        join_substrings.append(jmp_to_verify_while)

        clean_stack = [                     # removes all auxiliary values
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        while_body = (
            get_substring_left +
            get_substring_middle +
            get_substring_right +
            join_substrings
        )

        jmp_to_clean_stack = Opcode.get_jump_and_data(Opcode.JMPLE, get_bytes_count(while_body), True)
        verify_while[-1] = jmp_to_clean_stack

        return (
            initializing +
            verify_while +
            while_body +
            clean_stack
        )

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return LowerMethod(value)
        return super().build(value)



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.primitive.ibytestringtype import IByteStringType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class IsDigitMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            from boa3.model.type.primitive.bytestringtype import ByteStringType
            self_type = ByteStringType.build()

        identifier = 'isdigit'
        args: Dict[str, Variable] = {'self': Variable(self_type)}

        super().__init__(identifier, args, return_type=Type.bool)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.StackItem import StackItemType

        number0 = Integer(ord('0')).to_byte_array()
        number9 = Integer(ord('9')).to_byte_array()
        jmp_place_holder = (Opcode.JMP, b'\x01')

        initializing = [                    # initialize auxiliary values
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),              # index = len(string) - 1
            (Opcode.PUSH1, b''),            # isdigit = True
        ]

        verify_empty_string = [             # verifies if string is empty
            (Opcode.OVER, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder,               # jump to change_to_false if index == -1
        ]

        skip_first_verify_while = [         # skips the first while verification, since string is not empty
            jmp_place_holder
        ]

        verify_while = [                    # verifies if while is over
            (Opcode.OVER, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder,               # jump to return_bool if index >= len(string)
        ]

        jmp_verify_while = Opcode.get_jump_and_data(Opcode.JMP, get_bytes_count(verify_while), True)
        skip_first_verify_while[-1] = jmp_verify_while

        while_verify_lt_0 = [               # verifies if ord(string[index]) is < ord('0')
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH1, b''),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.DUP, b''),
            (Opcode.PUSHDATA1, Integer(len(number0)).to_byte_array() + number0),
            jmp_place_holder,               # if ord(string[index]) < ord('0'), return False
        ]

        while_verify_gt_9 = [               # verifies if ord(string[index]) is > ord('9')
            (Opcode.PUSHDATA1, Integer(len(number9)).to_byte_array() + number9),
            jmp_place_holder,               # if ord(string[index]) > ord('9'), return False
        ]

        while_go_to_verify = [              # decreases index and goes back to verify if there all characters were visited already
            (Opcode.SWAP, b''),
            (Opcode.DEC, b''),              # index--
            (Opcode.SWAP, b''),
            # jump back to verify_while
        ]

        jmp_back_to_verify = Opcode.get_jump_and_data(Opcode.JMP, -get_bytes_count(verify_while +
                                                                                   while_verify_lt_0 +
                                                                                   while_verify_gt_9 +
                                                                                   while_go_to_verify))
        while_go_to_verify.append(jmp_back_to_verify)

        jmp_out_of_while = Opcode.get_jump_and_data(Opcode.JMPLT, get_bytes_count(while_verify_gt_9 +
                                                                                  while_go_to_verify), True)
        while_verify_lt_0[-1] = jmp_out_of_while

        drop_char = [                       # remove extra char from stack
            (Opcode.DROP, b'')
        ]

        jmp_to_change_to_false = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(while_go_to_verify +
                                                                                        drop_char), True)
        while_verify_gt_9[-1] = jmp_to_change_to_false

        change_to_false = [                 # remove True on top of stack and put False
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),
        ]

        jmp_to_return = Opcode.get_jump_and_data(Opcode.JMPEQ, get_bytes_count(skip_first_verify_while +
                                                                               verify_while +
                                                                               while_verify_lt_0 +
                                                                               while_verify_gt_9 +
                                                                               while_go_to_verify +
                                                                               drop_char), True)
        verify_empty_string[-1] = jmp_to_return

        jmp_to_return = Opcode.get_jump_and_data(Opcode.JMPLT, get_bytes_count(while_verify_lt_0 +
                                                                               while_verify_gt_9 +
                                                                               while_go_to_verify +
                                                                               drop_char +
                                                                               change_to_false), True)
        verify_while[-1] = jmp_to_return

        clean_and_return_bool = [           # remove extra values from stack
            (Opcode.NIP, b''),
            (Opcode.NIP, b'')
        ]

        return (
            initializing +
            verify_empty_string +
            skip_first_verify_while +
            verify_while +
            while_verify_lt_0 +
            while_verify_gt_9 +
            while_go_to_verify +
            drop_char +
            change_to_false +
            clean_and_return_bool
        )

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return IsDigitMethod(value)
        return super().build(value)



from typing import Dict

from boa3.model.builtin.interop.interopevent import InteropEvent
from boa3.model.variable import Variable


class NotifyMethod(InteropEvent):

    def __init__(self):
        self._event_name_key = 'notification_name'

        from boa3.model.type.type import Type
        identifier = 'notify'
        syscall = 'System.Runtime.Notify'
        args: Dict[str, Variable] = {'state': Variable(Type.any),
                                     self._event_name_key: Variable(Type.str)
                                     }
        import ast
        event_name_default = ast.parse("'{0}'".format(identifier)
                                       ).body[0].value
        super().__init__(identifier, syscall, args, defaults=[event_name_default])

    @property
    def generate_name(self) -> bool:
        return False

    @property
    def args_to_generate(self) -> Dict[str, Variable]:
        return {key_name: value_type
                for key_name, value_type in self.args.items()
                if key_name != self._event_name_key}



def ceil(x: int, decimals: int) -> int:
    """
    Return the ceiling of x given the amount of decimals.
    This is the smallest integer >= x.

    :param x: any integer number
    :type x: int
    :param decimals: number of decimals
    :type decimals: int
    :return: the ceiling of x
    :rtype: int

    :raise Exception: raised when decimals is negative.
    """
    pass


def floor(x: int, decimals: int) -> int:
    """
    Return the floor of x given the amount of decimals.
    This is the largest integer <= x.

    :param x: any integer number
    :type x: int
    :param decimals: number of decimals
    :type decimals: int
    :return: the floor of x
    :rtype: int

    :raise Exception: raised when decimals is negative.
    """
    pass


def sqrt(x: int) -> int:
    """
    Gets the square root of a number.

    :param x: a non-negative number
    :type x: int
    :return: the square root of a number
    :rtype: int

    :raise Exception: raised when number is negative.
    """
    pass



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ReverseMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        if not isinstance(sequence_type, MutableSequenceType):
            from boa3.model.type.type import Type
            sequence_type = Type.mutableSequence

        identifier = 'reverse'
        args: Dict[str, Variable] = {'self': Variable(sequence_type)}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        return isinstance(params[0], IExpression) and isinstance(params[0].type, MutableSequenceType)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.REVERSEITEMS, b'')]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MutableSequenceType):
            return ReverseMethod(value)
        return super().build(value)



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class AbortMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'abort'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=Type.none)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.ABORT, b'')]

    @property
    def _args_on_stack(self) -> int:
        return 0

    @property
    def _body(self) -> Optional[str]:
        return None



from typing import Dict

from boa3.model.builtin.interop.blockchain.transactiontype import TransactionType
from boa3.model.builtin.interop.nativecontract import LedgerMethod
from boa3.model.variable import Variable


class GetTransactionFromBlockMethod(LedgerMethod):

    def __init__(self, transaction_type: TransactionType):
        from boa3.model.type.collection.sequence.uint256type import UInt256Type
        from boa3.model.type.type import Type

        identifier = 'get_transaction_from_block'
        syscall = 'getTransactionFromBlock'
        args: Dict[str, Variable] = {'block_hash_or_height': Variable(Type.union.build([UInt256Type.build(),
                                                                                        Type.int])),
                                     'tx_index': Variable(Type.int)}
        super().__init__(identifier, syscall, args, return_type=transaction_type)



from typing import List, Union

from boa3.builtin.interop.blockchain import Block, Signer, Transaction, VMState
from boa3.builtin.type import UInt256


class Ledger:
    """
    A class used to represent the Ledger native contract
    """

    @classmethod
    def get_block(cls, index_or_hash: Union[int, UInt256]) -> Block:
        """
        Gets the block with the given index or hash.

        :param index_or_hash: index or hash identifier of the block
        :type index_or_hash: int or UInt256
        :return: the desired block, if exists. None otherwise
        :rtype: Block or None
        """
        pass

    @classmethod
    def get_current_index(cls) -> int:
        """
        Gets the index of the current block.

        :return: the index of the current block
        :rtype: int
        """
        pass

    @classmethod
    def get_transaction(cls, hash_: UInt256) -> Transaction:
        """
        Gets a transaction with the given hash.

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: the Transaction, if exists. None otherwise
        """
        pass

    @classmethod
    def get_transaction_from_block(cls, block_hash_or_height: Union[UInt256, int], tx_index: int) -> Transaction:
        """
        Gets a transaction from a block.

        :param block_hash_or_height: a block identifier
        :type block_hash_or_height: UInt256 or int
        :param tx_index: the transaction identifier in the block
        :type tx_index: int
        :return: the Transaction, if exists. None otherwise
        """
        pass

    @classmethod
    def get_transaction_height(cls, hash_: UInt256) -> int:
        """
        Gets the height of a transaction.

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: height of the transaction
        """
        pass

    @classmethod
    def get_transaction_signers(cls, hash_: UInt256) -> List[Signer]:
        """
        Gets the VM state of a transaction.

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: VM state of the transaction
        """
        pass

    @classmethod
    def get_transaction_vm_state(cls, hash_: UInt256) -> VMState:
        """
        Gets the VM state of a transaction.

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: VM state of the transaction
        """
        pass



from abc import ABC
from typing import Iterable, Optional, Union


class CompilerWarning(ABC, BaseException):
    def __init__(self, line: int, col: int):
        self.line: int = line
        self.col: int = col
        self.filepath: Optional[str] = None

    @property
    def message(self) -> str:
        message = '' if self._warning_message is None else ' - ' + self._warning_message
        if isinstance(self.filepath, str):
            message += f'\t <{self.filepath}>'
        return '{0}:{1}{2}'.format(self.line, self.col, message)

    @property
    def _warning_message(self) -> Optional[str]:
        return None

    def __str__(self) -> str:
        return self.message

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.message == other.message


class InvalidArgument(CompilerWarning):
    """
    An warning raised when an attempt of method evaluation fails during optimization because an argument is invalid.
    """

    def __init__(self, line: int, col: int, custom_error_message: str = None):
        self.custom_error_message = custom_error_message
        super().__init__(line, col)

    @property
    def _warning_message(self) -> Optional[str]:
        message = "One or more arguments are invalid values"
        if self.custom_error_message is not None:
            message += f": {self.custom_error_message}"
        return message


class NameShadowing(CompilerWarning):
    """
    A warning raised when a name from an outer scope symbol is used as the name of an inner scope symbol
    """
    from boa3.model.symbol import ISymbol

    def __init__(self, line: int, col: int, outer_symbol: ISymbol, symbol_id: str):
        self.symbol_id: str = symbol_id
        self.existing_symbol = outer_symbol
        super().__init__(line, col)

    @property
    def _warning_message(self) -> Optional[str]:
        if self.symbol_id is not None:
            return "Shadowing {0} name '{1}'".format(self.existing_symbol.shadowing_name, self.symbol_id)


class RedeclaredSymbol(CompilerWarning):
    """
    A warning raised when a name from the same scope is used to identify multiple symbols
    """

    def __init__(self, line: int, col: int, symbol_id: str):
        self.symbol_id: str = symbol_id
        super().__init__(line, col)

    @property
    def _warning_message(self) -> Optional[str]:
        if self.symbol_id is not None:
            return "Redeclared '{0}' defined above".format(self.symbol_id)


class TypeCasting(CompilerWarning):
    """
    A warning raised when a type castings is used.
    """

    def __init__(self, line: int, col: int, origin_type_id: Union[str, Iterable[str]],
                 cast_type_id: Union[str, Iterable[str]]):
        if isinstance(origin_type_id, str):
            origin_type_id = [origin_type_id]
        if isinstance(cast_type_id, str):
            cast_type_id = [cast_type_id]

        self.origin_types = origin_type_id
        self.cast_types = cast_type_id
        super().__init__(line, col)

    @property
    def _warning_message(self) -> Optional[str]:
        origin_types = join_args(self.origin_types)
        cast_types = join_args(self.cast_types)
        return ("Casting {0} to {1}. Be aware that casting types may lead to runtime errors."
                .format(origin_types, cast_types)
                )


class UnreachableCode(CompilerWarning):
    """
    A warning raised when a block of code is detected as unreachable
    """

    def __init__(self, line: int, col: int):
        super().__init__(line, col)

    @property
    def _warning_message(self) -> Optional[str]:
        return "Unreachable code"


class UsingSpecificException(CompilerWarning):
    """
    A warning raised when a specific exception is used.
    """

    def __init__(self, line: int, col: int, exception_id: str):
        self._exception_id: str = exception_id
        super().__init__(line, col)

    @property
    def _warning_message(self) -> Optional[str]:
        return "{0} will be interpreted as BaseException when running in the blockchain".format(self._exception_id)


def join_args(iterable: Iterable[str]) -> str:
    return str.join("', '", iterable)



# for package imports in the smart contract
import boa3.builtin.interop.blockchain
import boa3.builtin.interop.contract
import boa3.builtin.interop.crypto
import boa3.builtin.interop.iterator
import boa3.builtin.interop.json
import boa3.builtin.interop.oracle
import boa3.builtin.interop.policy
import boa3.builtin.interop.role
import boa3.builtin.interop.runtime
import boa3.builtin.interop.stdlib
import boa3.builtin.interop.storage



from typing import List

from test_runner.blockchain.storage import TestRunnerStorage


class StorageCollection:
    def __init__(self):
        self._internal_list: List[TestRunnerStorage] = []

    def append(self, new_storage: TestRunnerStorage):
        return self._internal_list.append(new_storage)

    def remove(self, storage: TestRunnerStorage):
        return self._internal_list.remove(storage)

    def clear(self):
        return self._internal_list.clear()

    def __len__(self):
        return len(self._internal_list)

    def __contains__(self, item) -> bool:
        return any(storage.is_valid_identifier(item) for storage in self._internal_list)

    def __getitem__(self, item) -> TestRunnerStorage:
        return next((storage for storage in self._internal_list if storage.is_valid_identifier(item)), None)

    def __str__(self):
        dict_repr = {}
        for storage in self._internal_list:
            contract_id = storage.contract
            if contract_id is not None:
                contract_id = contract_id.name if hasattr(contract_id, 'name') else contract_id.script_hash

            dict_repr[contract_id] = storage.values
        return str(dict_repr)

    def __repr__(self):
        return str(self)



from enum import IntFlag


class FindOptions(IntFlag):
    NONE = 0

    KEYS_ONLY = 1 << 0
    REMOVE_PREFIX = 1 << 1
    VALUES_ONLY = 1 << 2
    DESERIALIZE_VALUES = 1 << 3
    PICK_FIELD_0 = 1 << 4
    PICK_FIELD_1 = 1 << 5



from typing import List, Tuple

from boa3.model.expression import IExpression
from boa3.model.symbol import ISymbol
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer


def get_bytes(instructions: List[Tuple[Opcode, bytes]]) -> bytes:
    return b''.join([opcode + arg for opcode, arg in instructions])


def get_bytes_count(instructions: List[Tuple[Opcode, bytes]]) -> int:
    return sum([len(opcode) + len(arg) for opcode, arg in instructions])


def get_storage_key_for_variable(symbol: ISymbol) -> bytes:
    if isinstance(symbol, IExpression):
        symbol_obj = symbol.origin
    else:
        symbol_obj = symbol

    origin_hash = Integer(hash(symbol_obj)).to_byte_array()
    symbol_hash = Integer(hash(symbol)).to_byte_array()
    return Integer(hash(origin_hash + symbol_hash)).to_byte_array()



from __future__ import annotations

from typing import Any, Dict, List, Optional

from boa3.neo import from_hex_str
from boa3.neo3.core.types import UInt256
from boa3_test.tests.test_classes.transaction import Transaction


class Block:
    def __init__(self, index: int):
        import time
        self._index: int = index
        # time() returns timestamp in nanoseconds and Neo uses timestamp in milliseconds
        self._timestamp: int = int(time.time_ns() / 1_000_000)
        self._hash: Optional[UInt256] = None
        self._transactions: List[Transaction] = []

    @property
    def index(self) -> int:
        return self._index

    @property
    def timestamp(self) -> int:
        return self._timestamp

    def get_transactions(self) -> List[Transaction]:
        """
        Gets a list of the block transactions. Changes in those transactions don't affect the ones inside the block.

        :return: the block transactions
        :rtype: List[Transaction]
        """
        return [tx.copy() for tx in self._transactions]

    def add_transaction(self, tx: Transaction):
        if all(block_tx != tx for block_tx in self._transactions):
            self._transactions.append(tx)

    @property
    def hash(self) -> Optional[bytes]:
        if self._hash is None:
            return None
        else:
            return self._hash.to_array()

    def to_json(self) -> Dict[str, Any]:
        return {
            'index': self._index,
            'timestamp': self._timestamp,
            'transactions': [tx.to_json() for tx in self._transactions]
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Block:
        # 'index' and 'timestamp' fields are required
        block = cls(int(json['index']))
        block._timestamp = int(json['timestamp'])

        if 'transactions' in json:
            tx_json = json['transactions']
            if not isinstance(tx_json, list):
                tx_json = [tx_json]
            block._transactions = [Transaction.from_json(js) for js in tx_json]

        if 'hash' in json and isinstance(json['hash'], str):
            block._hash = UInt256(from_hex_str(json['hash']))

        return block



from typing import Union

from boa3.builtin.interop.iterator import Iterator
from boa3.builtin.interop.storage.findoptions import FindOptions
from boa3.builtin.interop.storage.storagecontext import StorageContext
from boa3.builtin.interop.storage.storagemap import StorageMap
from boa3.builtin.type import ByteString


def get(key: ByteString, context: StorageContext = None) -> bytes:
    """
    Gets a value from the persistent store based on the given key.

    :param key: value identifier in the store
    :type key: str or bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: bytes
    """
    pass


def get_context() -> StorageContext:
    """
    Gets current storage context.

    :return: the current storage context
    :rtype: StorageContext
    """
    pass


def get_read_only_context() -> StorageContext:
    """
    Gets current read only storage context.

    :return: the current read only storage context
    :rtype: StorageContext
    """
    pass


def put(key: ByteString, value: Union[int, ByteString], context: StorageContext = None):
    """
    Inserts a given value in the key-value format into the persistent storage.

    :param key: the identifier in the store for the new value
    :type key: str or bytes
    :param value: value to be stored
    :type value: int or str or bytes
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def delete(key: ByteString, context: StorageContext = None):
    """
    Removes a given key from the persistent storage if exists.

    :param key: the identifier in the store for the new value
    :type key: str or bytes
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def find(prefix: ByteString,
         context: StorageContext = None,
         options: FindOptions = FindOptions.NONE) -> Iterator:
    """
    Searches in the storage for keys that start with the given prefix.

    :param prefix: prefix to find the storage keys
    :type prefix: str or bytes
    :param context: storage context to be used
    :type context: StorageContext
    :param options: the options of the search
    :type options: FindOptions
    :return: an iterator with the search results
    :rtype: Iterator
    """
    pass



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.collection.sequence.tupletype import TupleType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class MaxMethod(IBuiltinMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type
        identifier = 'max'

        self._allowed_types = [Type.int, Type.str, Type.bytes]
        default_type = Type.int
        if not self._is_valid_type(arg_value):
            arg_value = default_type

        args: Dict[str, Variable] = {
            'args1': Variable(arg_value),
            'args2': Variable(arg_value)
        }
        vararg = ('values', Variable(arg_value))
        super().__init__(identifier, args, return_type=arg_value, vararg=vararg)

    def _is_valid_type(self, arg_type: Optional[IType]) -> bool:
        return (isinstance(arg_type, IType) and
                any(allowed_type.is_type_of(arg_type) for allowed_type in self._allowed_types))

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_values.type is Type.int:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_values.type._identifier)

    @property
    def _arg_values(self) -> Variable:
        return self._vararg[1]

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.Integer import Integer

        jmp_place_holder = (Opcode.JMP, b'\x01')

        verify_number_of_parameters = [     # verifies if the stack has 2 or 3 items
            (Opcode.DEPTH, b''),
            (Opcode.PUSH2, b''),
            (Opcode.JMPEQ, b''),
            jmp_place_holder
        ]

        if_n_parameters_gt_2 = [         # if number of items in stack is 3 (2 ints and one tuple)
            (Opcode.REVERSE3, b''),
            (Opcode.UNPACK, b''),
            (Opcode.INC, b''),
            (Opcode.INC, b''),
            jmp_place_holder    # skips the next block of instructions
        ]

        if_n_parameters_eq_2 = [        # if number of items in stack is 2 (2 ints)
            (Opcode.PUSH2, b'')
        ]

        jmp_n_parameters_eq_2 = Opcode.get_jump_and_data(Opcode.JMP, get_bytes_count(if_n_parameters_eq_2), True)
        if_n_parameters_gt_2[-1] = jmp_n_parameters_eq_2

        jmp_n_parameters_gt_2 = Opcode.get_jump_and_data(Opcode.JMPEQ, get_bytes_count(if_n_parameters_gt_2))
        verify_number_of_parameters[-1] = jmp_n_parameters_gt_2

        repack_array = [        # pack all the arguments in the array
            (Opcode.PACK, b''),
        ]

        is_int_initialize = [   # puts the last array element as the max value
            (Opcode.DUP, b''),          # index = len(array) - 1
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),     # max = array[index]
        ]

        is_int_while = [    # this will get the next number in the array and compare it with the current max
            (Opcode.SWAP, b''),         # index--
            (Opcode.DEC, b''),
            (Opcode.SWAP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),         # max = max if max > array[index] else array[index]
            (Opcode.PICKITEM, b''),
        ]
        is_int_while.extend(self._compare_values())
        is_int_while.extend([
            (Opcode.OVER, b''),
            (Opcode.SIGN, b'')
            # if index != 0: go back to index--
            # else go to the end
        ])

        jmp_back_to_while_statement = (Opcode.JMPIF, Integer(-get_bytes_count(is_int_while)).to_byte_array(signed=True))
        is_int_while.append(jmp_back_to_while_statement)

        clean_stack = [    # removes everything but max
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        return (
            verify_number_of_parameters +
            if_n_parameters_gt_2 +
            if_n_parameters_eq_2 +
            repack_array +
            is_int_initialize +
            is_int_while +
            clean_stack
        )

    def _compare_values(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.MAX, b'')]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and len(value) > 0:
            value = value[0]
        if isinstance(value, TupleType):
            value = value.value_type
        if type(value) == type(self._arg_values.type):
            return self

        from boa3.model.builtin.method.maxbytestringmethod import MaxByteStringMethod
        from boa3.model.builtin.method.maxintmethod import MaxIntMethod
        from boa3.model.type.type import Type

        if Type.str.is_type_of(value) or Type.bytes.is_type_of(value):
            return MaxByteStringMethod(value)

        return MaxIntMethod(value)



from enum import IntFlag


class NamedCurve(IntFlag):
    """
    Represents the named curve used in ECDSA.
    """

    SECP256K1 = 22
    """
    The secp256k1 curve.

    :meta hide-value: 
    """

    SECP256R1 = 23
    """
    The secp256r1 curve, which known as prime256v1 or nistP-256.

    :meta hide-value:
    """



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class RemoveMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        if not isinstance(sequence_type, MutableSequenceType):
            from boa3.model.type.type import Type
            sequence_type = Type.mutableSequence

        self_arg = Variable(sequence_type)
        item_arg = Variable(sequence_type.value_type)

        identifier = 'remove'
        args: Dict[str, Variable] = {'self': self_arg,
                                     '__value': item_arg}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.model.type.itype import IType
        sequence_type: IType = params[0].type
        value_type: IType = params[1].type

        if not isinstance(sequence_type, MutableSequenceType):
            return False
        return sequence_type.value_type.is_type_of(value_type)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        return [
            (Opcode.PUSH0, b''),  # need to find the index to use REMOVE opcode
            (Opcode.DUP, b''),      # while index < len(array):
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.GE, b''),
            (Opcode.JMPIF, Integer(13).to_byte_array(signed=True, min_length=1)),
            (Opcode.PUSH2, b''),        # if array[index] == value
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),             # break
            (Opcode.JMPEQ, Integer(5).to_byte_array(signed=True, min_length=1)),
            (Opcode.INC, b''),          # index = index + 1
            (Opcode.JMP, Integer(-16).to_byte_array(signed=True, min_length=1)),
            (Opcode.NIP, b''),
            (Opcode.REMOVE, b''),
        ]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MutableSequenceType):
            return RemoveMethod(value)
        return super().build(value)



from __future__ import annotations

import base64
from typing import Any, Dict


class Witness:
    def __init__(self, invocation_script: bytes, verification_script: bytes):
        self._invocation_script = invocation_script
        self._verification_script = verification_script

    @property
    def invocation_script(self) -> bytes:
        return self._invocation_script

    @property
    def verification_script(self) -> bytes:
        return self._verification_script

    def to_json(self) -> Dict[str, Any]:
        import base64
        return {
            'invocation': base64.b64encode(self._invocation_script).decode(),
            'verification': base64.b64encode(self._verification_script).decode()
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Witness:
        invocation = base64.b64decode(json['invocation'])
        verification = base64.b64decode(json['verification'])
        return cls(invocation, verification)



from typing import Any, Dict, Optional, Sequence

from boa3.neo.core.types.InteropInterface import InteropInterface
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItem import StackItemType
from boa3.neo.vm.type.String import String


def stack_item_from_json(item: Dict[str, Any]) -> Any:
    if 'type' not in item:
        return None

    item_type: StackItemType = StackItemType.get_stack_item_type(item['type'])
    if item_type is StackItemType.InteropInterface:
        if 'iterator' in item:
            return [stack_item_from_json(value) for value in item['iterator']]
        else:
            return InteropInterface
    if item_type is StackItemType.Any or 'value' not in item:
        return None

    value: Any = None
    item_value: Any = item['value']

    if item_type is StackItemType.Boolean:
        if isinstance(item_value, str) and item_value in (str(True), str(False)):
            item_value = item_value == str(True)
        if not isinstance(item_value, bool):
            raise ValueError
        value = item_value

    elif item_type is StackItemType.Integer:
        if isinstance(item_value, str):
            item_value = int(item_value)
        if not isinstance(item_value, int):
            raise ValueError
        value = item_value

    elif item_type in (StackItemType.ByteString, StackItemType.Buffer):
        if not isinstance(item_value, str):
            raise ValueError

        import base64
        decoded: bytes = base64.b64decode(item_value)

        try:
            value = String.from_bytes(decoded)
        except BaseException:
            value = decoded

    elif item_type in (StackItemType.Array, StackItemType.Struct):
        if not isinstance(item_value, Sequence) or isinstance(item_value, (str, bytes)):
            raise ValueError
        value = [stack_item_from_json(x) for x in item_value]

    elif item_type is StackItemType.Map:
        if not isinstance(item_value, Sequence):
            raise ValueError
        value = {}

        for x in item_value:
            if 'key' not in x or 'value' not in x:
                raise ValueError
            value[stack_item_from_json(x['key'])] = stack_item_from_json(x['value'])

    return value


def bytes_from_json(item: Dict[str, Any]) -> Optional[bytes]:
    value = stack_item_from_json(item)

    if isinstance(value, str):
        value = String(value).to_bytes()
    elif isinstance(value, int):
        value = Integer(value).to_byte_array()

    return value if isinstance(value, bytes) else None


def contract_parameter_to_json(value: Any) -> Dict[str, Any]:
    if value is None:
        return {'type': AbiType.Any}

    stack_type: AbiType = AbiType.Any
    parameter_value: Any = None

    if isinstance(value, bool):
        stack_type = AbiType.Boolean
        parameter_value = value
    elif isinstance(value, int):
        stack_type = AbiType.Integer
        parameter_value = value
    elif isinstance(value, str):
        stack_type = AbiType.String
        parameter_value = value
    elif isinstance(value, (bytes, bytearray)):
        import base64
        stack_type = AbiType.ByteArray
        parameter_value = String.from_bytes(base64.b64encode(value))
    elif isinstance(value, Sequence):
        stack_type = AbiType.Array
        parameter_value = [contract_parameter_to_json(x) for x in value]
    elif isinstance(value, dict):
        stack_type = AbiType.Map
        parameter_value = [{'key': contract_parameter_to_json(key),
                            'value': contract_parameter_to_json(value)
                            } for key, value in value.items()
                           ]

    result = {'type': stack_type.value}
    if parameter_value is not None:
        result['value'] = parameter_value

    return result



from boa3.builtin.type import UInt160


class Notification:
    """
    Represents a notification.

    :ivar script_hash: the script hash of the notification sender
    :vartype script_hash: UInt160
    :ivar event_name: the notification's name
    :vartype event_name: str
    :ivar state: a tuple value storing all the notification contents.
    :vartype state: tuple
    """

    def __init__(self):
        self.script_hash: UInt160 = UInt160()
        self.event_name: str = ''
        self.state: tuple = ()



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class BoolMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'bool'

        args: Dict[str, Variable] = {
            'value': Variable(Type.any),
        }
        super().__init__(identifier, args, return_type=Type.bool)

    @property
    def _value(self) -> Variable:
        return self.args['value']

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.StackItem import StackItemType
        from boa3.compiler.codegenerator import get_bytes_count
        jmp_place_holder = (Opcode.JMP, b'\x01')

        verify_is_none = [  # verify if value is None
            (Opcode.DUP, b''),
            (Opcode.ISNULL, b''),
            jmp_place_holder
        ]

        value_is_none = [   # if value is None, then return False
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder
        ]

        verify_is_array = [     # if value is an array, check the length of it
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, StackItemType.Array),
            jmp_place_holder
        ]

        verify_is_map = [       # if value is a map, check the length of it
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, StackItemType.Map),
            jmp_place_holder
        ]

        verify_is_int = [       # if value is an int, check if it is non zero
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, StackItemType.Integer),
            jmp_place_holder
        ]

        verify_is_bytestring = [    # if value is a bytestring, check if it is non zero
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, StackItemType.ByteString),
            jmp_place_holder
        ]

        put_true = [    # if value it's not a type from above, return True
            (Opcode.DROP, b''),
            (Opcode.PUSH1, b''),
            jmp_place_holder
        ]

        collection_non_zero = [     # get length of collection
            (Opcode.SIZE, b''),
        ]

        non_zero = [        # verify if it's non zero
            (Opcode.NZ, b''),
        ]

        # region jump logic

        jump_instructions = collection_non_zero + non_zero
        put_true[-1] = Opcode.get_jump_and_data(Opcode.JMP, get_bytes_count(jump_instructions), True)

        jump_instructions = collection_non_zero + put_true
        verify_is_bytestring[-1] = Opcode.get_jump_and_data(Opcode.JMPIF, get_bytes_count(jump_instructions), True)

        jump_instructions = collection_non_zero + put_true + verify_is_bytestring
        verify_is_int[-1] = Opcode.get_jump_and_data(Opcode.JMPIF, get_bytes_count(jump_instructions), True)

        jump_instructions = put_true + verify_is_int + verify_is_bytestring
        verify_is_map[-1] = Opcode.get_jump_and_data(Opcode.JMPIF, get_bytes_count(jump_instructions), True)

        jump_instructions = verify_is_map + verify_is_int + verify_is_bytestring + put_true
        verify_is_array[-1] = Opcode.get_jump_and_data(Opcode.JMPIF, get_bytes_count(jump_instructions), True)

        jump_instructions = (
            verify_is_array +
            verify_is_map +
            verify_is_int +
            verify_is_bytestring +
            put_true +
            collection_non_zero +
            non_zero
        )
        value_is_none[-1] = Opcode.get_jump_and_data(Opcode.JMP, get_bytes_count(jump_instructions), True)

        jump_instructions = value_is_none
        verify_is_none[-1] = Opcode.get_jump_and_data(Opcode.JMPIFNOT, get_bytes_count(jump_instructions), True)

        # endregion

        bool_method = (
            verify_is_none +
            value_is_none +
            verify_is_array +
            verify_is_map +
            verify_is_int +
            verify_is_bytestring +
            put_true +
            collection_non_zero +
            non_zero
        )

        return bool_method

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



from typing import Dict, List, Optional, Tuple

from boa3.constants import POLICY_SCRIPT
from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class GetPolicyContractScriptHashMethod(IBuiltinMethod):
    def __init__(self):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        identifier = '-get_policy_contract'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=UInt160Type.build())

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer

        value = POLICY_SCRIPT
        return [
            (Opcode.PUSHDATA1, Integer(len(value)).to_byte_array() + value)
        ]


class PolicyContractProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'PolicyContract'
        getter = GetPolicyContractScriptHashMethod()
        super().__init__(identifier, getter)


PolicyContract = PolicyContractProperty()



from __future__ import annotations

from typing import Any, Dict, Tuple

from boa3.neo import to_hex_str
from boa3.neo.utils import stack_item_from_json


class Notification:
    _event_name_key = 'eventname'
    _script_hash_key = 'scripthash'
    _value_key = 'value'

    def __init__(self, event_name: str, script_hash: bytes, *value: Any):
        self._event_name: str = event_name
        self._script_hash: bytes = script_hash
        self._value: Tuple[Any] = value

    @property
    def name(self) -> str:
        return self._event_name

    @property
    def origin(self) -> bytes:
        return self._script_hash

    @property
    def arguments(self) -> tuple:
        return self._value

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Notification:
        """
        Creates a Notification object from a json.

        :param json: json that contains the notification data
        :return: a Notification object
        :rtype: Notification
        """
        keys = set(json.keys())
        if not keys.issubset([cls._event_name_key, cls._script_hash_key, cls._value_key]):
            return None

        name: str = json[cls._event_name_key] if cls._event_name_key in json else ""
        script: bytes = json[cls._script_hash_key] if cls._script_hash_key in json else b''
        try:
            value: Any = stack_item_from_json(json[cls._value_key]) if cls._value_key in json else []
        except ValueError:
            value = []

        if not isinstance(value, list):
            value = [value]

        if isinstance(script, str):
            script = cls._get_script_from_str(script)

        return cls(name, script, *value)

    @classmethod
    def _get_script_from_str(cls, script: str) -> bytes:
        if isinstance(script, str) and script.startswith('0x'):
            bytes_script = bytearray()
            for x in range(2, len(script), 2):
                bytes_script.append(int(script[x:x + 2], 16))
            bytes_script.reverse()
            script = bytes(bytes_script)
        return script

    def __str__(self) -> str:
        return '[{0}] {1}'.format(to_hex_str(self.origin), self._event_name)

    def __repr__(self) -> str:
        return str(self)



from typing import Union


def to_script_hash(data_bytes: bytes) -> bytes:
    """
    Converts a data to a script hash.

    :param data_bytes: data to hash.
    :type data_bytes: bytearray or bytes

    :return: the scripthash of the data
    :rtype: bytes
    """
    from boa3.neo import cryptography
    from base58 import b58decode
    try:
        base58_decoded = b58decode(data_bytes)[1:]  # first byte is the address version

        from boa3.constants import SIZE_OF_INT160
        return bytes(base58_decoded[:SIZE_OF_INT160])
    except BaseException:
        return cryptography.hash160(data_bytes)


def public_key_to_script_hash(public_key: Union[str, bytes]) -> bytes:
    """
    Converts a public key bytes sequence to a script hash.

    :return: the script hash of the data
    :rtype: bytes
    """
    if isinstance(public_key, str):
        public_key = from_hex_str(public_key)

    from boa3.constants import SIZE_OF_ECPOINT

    if len(public_key) != SIZE_OF_ECPOINT:
        # it's not a public key
        return to_script_hash(public_key)

    # public keys must include a check sig script when converting to script hash
    from boa3.neo.vm.opcode.Opcode import Opcode
    from boa3.neo.vm.type.Integer import Integer
    from boa3.model.builtin.interop.crypto.checksigmethod import CheckSigMethod

    check_sig_arg = Opcode.PUSHDATA1 + Integer(SIZE_OF_ECPOINT).to_byte_array() + public_key
    check_sig_call = CheckSigMethod.get_raw_bytes()
    return to_script_hash(check_sig_arg + check_sig_call)


def to_hex_str(data_bytes: bytes) -> str:
    """
    Converts bytes into its string hex representation.

    :param data_bytes: data to represent as hex.
    :type data_bytes: bytearray or bytes

    :return: the hex representation of the data
    :rtype: str
    """
    if isinstance(data_bytes, bytes):
        data_bytes = bytearray(data_bytes)
    data_copy = data_bytes.copy()
    data_copy.reverse()
    return '0x' + data_copy.hex()


def from_hex_str(hex_string: str) -> bytes:
    """
    Converts a string hex representation to the equivalent bytes.

    :param hex_string: data hex representation.
    :type hex_string: str

    :return: the represented bytes
    :rtype: bytes
    """
    if hex_string.startswith('0x'):
        hex_string = hex_string[2:]

    data_bytes = bytearray.fromhex(hex_string)
    data_bytes.reverse()
    return bytes(data_bytes)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

from boa3.model.operation.operator import Operator
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode.Opcode import Opcode


class IOperation(ISymbol, ABC):
    """
    An interface used to represent operations

    :ivar operator: the operator of the operation
    :ivar result: the result type of the operation
    """

    def __init__(self, operator: Operator, result_type: IType):
        self.operator: Operator = operator
        self.result: IType = result_type

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        """
        Gets the operation sequence of opcodes with in Neo Vm

        :return: the opcode if exists. Empty list otherwise.
        """
        return []

    @property
    def shadowing_name(self) -> str:
        """
        Gets the type of the evaluated expression

        :return: the resulting type when the expression is evaluated
        """
        return self.operator.value

    @property
    def bytecode(self) -> bytes:
        """
        Gets the operation bytecode

        :return: the bytecode if exists. Empty bytes otherwise.
        """
        str_mult_bytes = bytearray()
        for opcode, data in self.opcode:
            str_mult_bytes += opcode + data
        return bytes(str_mult_bytes)

    @property
    @abstractmethod
    def number_of_operands(self) -> int:
        """
        Gets the number of operands required for this operations

        :return: Number of operands
        """
        pass

    @property
    def op_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments. Same from `number_of_operands` by default.
        """
        return self.number_of_operands

    @abstractmethod
    def validate_type(self, *types: IType) -> bool:
        """
        Verifies if the given operands are valid to the operation

        :param types: types of the operand
        :return: True if all arguments are valid. False otherwise.
        """
        pass

    def is_valid(self, operator: Operator, *types: IType) -> bool:
        """
        Verifies if the given operator and operands are valid to the operation

        :param operator:
        :param types: types of the operand
        :return: True if all arguments are valid. False otherwise.
        """
        if len(types) != self.number_of_operands:
            return False

        return operator is self.operator and self.validate_type(*types)

    @property
    def is_supported(self) -> bool:
        """
        Verifies if the operation is supported by the compiler

        :return: True if it is supported. False otherwise.
        """
        return True

    @classmethod
    @abstractmethod
    def build(cls, *operands: IType) -> Optional[IOperation]:
        """
        Creates an operation with the given operands types

        :param operands: operands types
        :return: The built operation if the operands are valid. None otherwise
        :rtype: IOperation or None
        """
        return None



from __future__ import annotations

import ast
from typing import Dict, List, Optional

from boa3 import constants
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.analyser.astoptimizer import AstOptimizer
from boa3.analyser.constructanalyser import ConstructAnalyser
from boa3.analyser.moduleanalyser import ModuleAnalyser
from boa3.analyser.supportedstandard.standardanalyser import StandardAnalyser
from boa3.analyser.typeanalyser import TypeAnalyser
from boa3.builtin import NeoMetadata
from boa3.exception.CompilerError import CompilerError
from boa3.exception.CompilerWarning import CompilerWarning
from boa3.model.symbol import ISymbol
from boa3.model.type.type import Type


class Analyser:
    """
    This class is responsible for the semantic analysis of the code

    :ivar symbol_table: a dictionary used to store the identifiers
    """

    def __init__(self, ast_tree: ast.AST, path: str = None, project_root: str = None, log: bool = False):
        self.symbol_table: Dict[str, ISymbol] = {}

        self.ast_tree: ast.AST = ast_tree
        self.metadata: NeoMetadata = NeoMetadata()
        self.is_analysed: bool = False
        self._log: bool = log

        self.__include_builtins_symbols()
        self._errors = []
        self._warnings = []
        self._imported_files: Dict[str, Analyser] = {}
        self._included_imported_files: bool = False

        import os
        self.path: str = path
        self.filename: str = path if path is None else os.path.realpath(path)

        if project_root is not None:
            if not os.path.exists(project_root):
                project_root = os.path.abspath(f'{os.path.curdir}{os.path.sep}{project_root}')

            if os.path.isfile(project_root):
                project_root = os.path.dirname(os.path.abspath(project_root))

        self.root: str = (os.path.realpath(project_root)
                          if project_root is not None and os.path.isdir(project_root)
                          else path)

    @staticmethod
    def analyse(path: str, log: bool = False,
                imported_files: Optional[Dict[str, Analyser]] = None,
                import_stack: Optional[List[str]] = None,
                root: str = None) -> Analyser:
        """
        Analyses the syntax of the Python code

        :param path: the path of the Python file
        :param log: if compiler errors should be logged.
        :param import_stack: a list that represents the current import stack if it's from an import.
                             If it's not triggered by an import, must be None.
        :param imported_files: a dict that maps the paths of the files that were analysed if it's from an import.
                               If it's not triggered by an import, must be None.
        :param root: the path of the project root that the current smart contract is part of.
        :return: a boolean value that represents if the analysis was successful
        :rtype: Analyser
        """
        with open(path, 'rb') as source:
            ast_tree = ast.parse(source.read())

        analyser = Analyser(ast_tree, path, root if isinstance(root, str) else path, log)
        analyser.__pre_execute()

        # fill symbol table
        if not analyser.__analyse_modules(imported_files, import_stack):
            return analyser
        # check if standards are correctly implemented
        if not analyser.__check_standards():
            return analyser
        # check is the types are correct
        if not analyser.__check_types():
            return analyser

        analyser.__pos_execute()
        analyser.is_analysed = True

        return analyser

    @property
    def errors(self) -> List[CompilerError]:
        return self._errors.copy()

    @property
    def warnings(self) -> List[CompilerWarning]:
        return self._warnings.copy()

    def copy(self) -> Analyser:
        copied = Analyser(ast_tree=self.ast_tree, path=self.path, project_root=self.root, log=self._log)

        copied.metadata = self.metadata
        copied.is_analysed = self.is_analysed
        copied.symbol_table = self.symbol_table.copy()
        copied.filename = self.filename
        copied._imported_files = self._imported_files.copy()
        copied._included_imported_files = self._included_imported_files

        return copied

    def __include_builtins_symbols(self):
        """
        Include the Python builtins in the global symbol table
        """
        self.symbol_table.update(Type.builtin_types())

    def __check_types(self) -> bool:
        """
        Performs the type checking

        :return: a boolean value that represents if the analysis was successful
        """
        type_analyser = TypeAnalyser(self, self.symbol_table, log=self._log)
        self.__update_logs(type_analyser)
        return not type_analyser.has_errors

    def __analyse_modules(self,
                          imported_files: Optional[Dict[str, Analyser]] = None,
                          import_stack: Optional[List[str]] = None) -> bool:
        """
        Validates the symbols and constructs the symbol table of the ast tree

        :return: a boolean value that represents if the analysis was successful
        """
        module_analyser = ModuleAnalyser(self, self.symbol_table,
                                         log=self._log,
                                         filename=self.filename,
                                         root_folder=self.root,
                                         analysed_files=imported_files,
                                         import_stack=import_stack)
        self.symbol_table.update(module_analyser.global_symbols)
        self.ast_tree.body.extend(module_analyser.imported_nodes)
        self.__update_logs(module_analyser)
        self._imported_files = module_analyser.analysed_files.copy()
        return not module_analyser.has_errors

    def __check_standards(self) -> bool:
        """
        Verify if the standards included in the metadata are fully implemented

        :return: a boolean value that represents if the analysis was successful
        """
        standards_analyser = StandardAnalyser(self, self.symbol_table, log=self._log)
        self.__update_logs(standards_analyser)
        return not standards_analyser.has_errors

    def __update_logs(self, analyser: IAstAnalyser):
        self._errors.extend(analyser.errors)
        self._warnings.extend(analyser.warnings)

    def __pre_execute(self):
        """
        Pre executes the instructions of the ast for optimization
        """
        self.ast_tree = ConstructAnalyser(self.ast_tree, log=self._log).tree

    def __pos_execute(self):
        """
        Tries to optimize the ast after validations
        """
        optimizer = AstOptimizer(self, log=self._log)
        self.__update_logs(optimizer)

    def update_symbol_table(self, symbol_table: Dict[str, ISymbol]):
        for symbol_id, symbol in symbol_table.items():
            if (hasattr(symbol, 'origin')
                    and hasattr(symbol.origin, 'origin')
                    and isinstance(symbol.origin.origin, ast.AST)
                    and len(symbol_id.split(constants.VARIABLE_NAME_SEPARATOR)) <= 1):

                if symbol_id in self.symbol_table:
                    self.symbol_table.pop(symbol_id)

                origin_hash = symbol.origin.origin.__hash__()
                unique_id = '{0}{2}{1}'.format(origin_hash, symbol_id, constants.VARIABLE_NAME_SEPARATOR)
            else:
                unique_id = symbol_id

            from boa3.model.identifiedsymbol import IdentifiedSymbol
            from boa3.model.type.classes.userclass import UserClass
            if not isinstance(symbol, IdentifiedSymbol) or isinstance(symbol, UserClass):
                if unique_id not in self.symbol_table:
                    self.symbol_table[unique_id] = symbol

    def update_symbol_table_with_imports(self):
        if self._included_imported_files:
            return

        import os
        from boa3.analyser.importanalyser import ImportAnalyser
        from boa3.model.imports.importsymbol import Import

        imports = self._imported_files.copy()
        paths_already_imported = [imported.origin for imported in self.symbol_table.values()
                                  if isinstance(imported, Import) and isinstance(imported.origin, str)]
        if isinstance(self.path, str):
            paths_already_imported.append(self.path.replace(os.path.sep, constants.PATH_SEPARATOR))

        for file_path, analyser in imports.items():
            if file_path not in paths_already_imported:
                import_analyser = ImportAnalyser(file_path, self.root,
                                                 already_imported_modules=imports,
                                                 log=False, get_entry=True)
                import_symbol = Import(file_path, analyser.ast_tree, import_analyser, {})
                self.symbol_table[file_path] = import_symbol

        self._included_imported_files = True



from typing import Any, Dict, List, Optional, Tuple

from boa3.compiler.codegenerator import get_bytes_count
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ECPointMethod(IBuiltinMethod):

    def __init__(self, return_type: ECPointType, argument_type: IType = None):
        from boa3.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.none

        identifier = 'ECPoint'
        args: Dict[str, Variable] = {'arg': Variable(argument_type)}

        super().__init__(identifier, args, return_type=return_type)

    @property
    def _arg_arg(self) -> Variable:
        return self.args['arg']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_arg.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_arg.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.model.type.type import Type
        return Type.bytes.is_type_of(param_type)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.StackItem import StackItemType

        ECPOINT_SIZE = 33

        throw_if_invalid = [
            (Opcode.THROW, b''),
        ]
        check_bytestr_size = [
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSHINT8, Integer(ECPOINT_SIZE).to_byte_array(signed=True)),
            Opcode.get_jump_and_data(Opcode.JMPEQ, get_bytes_count(throw_if_invalid), jump_through=True),
        ]

        return [
            (Opcode.CONVERT, StackItemType.ByteString),  # convert to ECPoint
            (Opcode.DUP, b''),
            (Opcode.ISNULL, b''),
            Opcode.get_jump_and_data(Opcode.JMPIF, get_bytes_count(check_bytestr_size), jump_through=True),
        ] + check_bytestr_size + throw_if_invalid

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_arg.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return ECPointMethod(self.return_type, value)
        return super().build(value)



from __future__ import annotations

from typing import Any, Dict, Optional

from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.type.classes.classstructtype import ClassStructType
from boa3.model.variable import Variable


class ContractAbiType(ClassStructType):
    """
    A class used to represent Neo ContractAbi class
    """

    def __init__(self):
        super().__init__('ContractAbi')
        from boa3.model.builtin.interop.contract.contractmanifest.contractmethoddescriptortype import \
            ContractMethodDescriptorType
        from boa3.model.builtin.interop.contract.contractmanifest.contracteventdescriptortype import \
            ContractEventDescriptorType
        from boa3.model.type.type import Type

        self._variables: Dict[str, Variable] = {
            'methods': Variable(Type.list.build_collection(ContractMethodDescriptorType.build())),
            'events': Variable(Type.list.build_collection(ContractEventDescriptorType.build()))
        }
        self._constructor: Method = None

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> Dict[str, Method]:
        return {}

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> ContractAbiType:
        if value is None or cls._is_type_of(value):
            return _ContractAbi

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, ContractAbiType)


_ContractAbi = ContractAbiType()



from __future__ import annotations

import base64
from typing import Any, Dict, List, Optional

from boa3.neo import from_hex_str
from boa3.neo3.core.types import UInt256
from boa3.neo3.vm import VMState, vmstate
from boa3_test.tests.test_classes import transactionattribute as tx_attribute
from boa3_test.tests.test_classes.signer import Signer
from boa3_test.tests.test_classes.witness import Witness


class Transaction:
    def __init__(self, script: bytes, signers: List[Signer] = None, witnesses: List[Witness] = None):
        self._signers: List[Signer] = signers if signers is not None else []
        self._witnesses: List[Witness] = witnesses if witnesses is not None else []
        self._attributes: List[tx_attribute.TransactionAttribute] = []
        self._script: bytes = script
        self._hash: Optional[UInt256] = None
        self._state: VMState = VMState.NONE

    def add_attribute(self, tx_attr: tx_attribute.TransactionAttribute):
        if tx_attr not in self._attributes:
            self._attributes.append(tx_attr)

    @property
    def hash(self) -> Optional[bytes]:
        if self._hash is None:
            return None
        else:
            return self._hash.to_array()

    @property
    def state(self) -> VMState:
        return self._state

    def to_json(self) -> Dict[str, Any]:
        from boa3.neo.vm.type.String import String
        return {
            'signers': [signer.to_json() for signer in self._signers],
            'witnesses': [witness.to_json() for witness in self._witnesses],
            'attributes': [attr.to_json() for attr in self._attributes],
            'script': String.from_bytes(base64.b64encode(self._script)),
            'state': self._state.name
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Transaction:
        script = base64.b64decode(json['script'])
        tx = cls(script)
        if 'signers' in json:
            signers_json = json['signers']
            if not isinstance(signers_json, list):
                signers_json = [signers_json]
            tx._signers = [Signer.from_json(js) for js in signers_json]

        if 'witnesses' in json:
            witnesses_json = json['witnesses']
            if not isinstance(witnesses_json, list):
                witnesses_json = [witnesses_json]
            tx._witnesses = [Witness.from_json(js) for js in witnesses_json]

        if 'attributes' in json:
            attributes_json = json['attributes']
            if not isinstance(attributes_json, list):
                attributes_json = [attributes_json]
            tx._attributes = [tx_attribute.TransactionAttribute.from_json(js) for js in attributes_json]

        if 'hash' in json and isinstance(json['hash'], str):
            tx._hash = UInt256(from_hex_str(json['hash']))

        if 'state' in json and isinstance(json['state'], str):
            tx._state = vmstate.get_vm_state(json['state'])

        return tx

    def copy(self):
        copied = Transaction(self._script, self._signers, self._witnesses)
        copied._hash = self._hash
        return copied

    def __eq__(self, other) -> bool:
        if not isinstance(other, Transaction):
            return False
        if self._hash == other._hash:
            return True

        return (self._script == other._script
                and self._attributes == self._attributes
                and self._signers == other._signers
                and self._witnesses == other._witnesses)



from typing import Any, List, Optional, Tuple

from boa3 import constants
from boa3.model.method import Method
from boa3.model.type.itype import IType
from boa3.model.type.primitive.bytestype import BytesType
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.AbiType import AbiType


class ECPointType(BytesType):
    """
    A class used to represent Neo's UInt160 type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'ECPoint'
        from boa3.model.builtin.method.ecpointmethod import ECPointMethod
        self._constructor = ECPointMethod(self)

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def abi_type(self) -> AbiType:
        return AbiType.PublicKey

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @property
    def default_value(self) -> Any:
        return bytes(constants.SIZE_OF_ECPOINT)

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _ECPoint

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, ECPointType)

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.model.builtin.builtin import Builtin

        instance_methods = [Builtin.ScriptHash,
                            ]

        for instance_method in instance_methods:
            self._instance_methods[instance_method.raw_identifier] = instance_method.build(self)

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.type.classes.pythonclass import PythonClass
        return super(PythonClass, self).is_instance_opcodes()

    def _is_instance_inner_opcodes(self, jmp_to_if_false: int = 0) -> List[Tuple[Opcode, bytes]]:
        from boa3 import constants
        push_int_opcode, size_data = Opcode.get_push_and_data(constants.SIZE_OF_ECPOINT)

        return [
            (Opcode.SIZE, b''),  # return len(value) == 33
            (push_int_opcode, size_data),
            (Opcode.NUMEQUAL, b'')
        ]


_ECPoint = ECPointType()



from typing import Dict

from boa3.model.builtin.interop.contract.contracttype import ContractType
from boa3.model.builtin.interop.nativecontract.ContractManagement.contractmanagementmethod import \
    ContractManagementMethod
from boa3.model.variable import Variable


class GetContractMethod(ContractManagementMethod):

    def __init__(self, contract_type: ContractType):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        identifier = 'get_contract'
        syscall = 'getContract'
        args: Dict[str, Variable] = {'hash': Variable(UInt160Type.build())}
        super().__init__(identifier, syscall, args, return_type=contract_type)



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.mapping.mutable.dicttype import DictType
from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class PopMethod(IBuiltinMethod):
    def __init__(self, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        identifier = 'pop'

        super().__init__(identifier, args, defaults=defaults, return_type=return_type)

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type

        if self._arg_self.type is Type.mutableSequence:
            return self._identifier

        if Type.dict.is_type_of(self._arg_self.type):

            if len(self.args) == 2:
                return '-{0}_{1}'.format(self._identifier, Type.dict.identifier)

            elif len(self.args) == 3:
                return '-{0}_{1}_{2}'.format(self._identifier, Type.dict.identifier, 'default')

        if Type.mutableSequence.is_type_of(self._arg_self.type):
            return '-{0}_{1}'.format(self._identifier, Type.mutableSequence.identifier)

        return self._identifier

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) < 0 or len(params) > 2:
            return False

        if any(not isinstance(param, (IExpression, IType)) for param in params):
            return False

        sequence_or_map = params[0].type if isinstance(params[0], IExpression) else params[0]
        if not isinstance(sequence_or_map, MutableSequenceType) or not isinstance(sequence_or_map, DictType):
            return False

        if len(params) > 1:
            value = params[1].type if isinstance(params[1], IExpression) else params[1]
            return sequence_or_map.valid_key.is_type_of(value)
        return True

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        return [
            (Opcode.DUP, b''),
            (Opcode.SIGN, b''),
            (Opcode.PUSHM1, b''),
            (Opcode.JMPNE, Integer(5).to_byte_array(min_length=1, signed=True)),
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.SWAP, b''),
            (Opcode.REMOVE, b''),
        ]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        from boa3.model.builtin.classmethod.popdictmethod import PopDictMethod
        from boa3.model.builtin.classmethod.popsequencemethod import PopSequenceMethod
        from boa3.model.type.type import Type

        if Type.dict.is_type_of(value[0]):
            if len(value) == 3:
                from boa3.model.builtin.classmethod.popdictdefaultmethod import PopDictDefaultMethod
                return PopDictDefaultMethod(value[0])
            return PopDictMethod(value[0])

        return PopSequenceMethod(value[0])



import os
import platform
import sys

from boa3 import __version__ as boa_version
from boa3.neo import from_hex_str

ENCODING = 'utf-8'
BYTEORDER = 'little'

ATTRIBUTE_NAME_SEPARATOR = '.'
VARIABLE_NAME_SEPARATOR = ','
PATH_SEPARATOR = '/'
IMPORT_WILDCARD = '*'

SYS_VERSION_INFO = sys.version_info
SYS_VERSION = platform.python_version()
BOA_VERSION = boa_version
BOA_PACKAGE_PATH = os.path.abspath(f'{__file__}/..')

ONE_BYTE_MAX_VALUE = 255
TWO_BYTES_MAX_VALUE = 256 ** 2 - 1
FOUR_BYTES_MAX_VALUE = 256 ** 4 - 1

SIZE_OF_INT32 = 4
SIZE_OF_INT160 = 20
SIZE_OF_INT256 = 32
SIZE_OF_ECPOINT = 33
DEFAULT_UINT32 = 0
GAS_DECIMALS = 8

INIT_METHOD_ID = '__init__'
INITIALIZE_METHOD_ID = '_initialize'
DEPLOY_METHOD_ID = '_deploy'

NEO_SCRIPT = from_hex_str('0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5')
GAS_SCRIPT = from_hex_str('0xd2a4cff31913016155e38e474a2c06d08be276cf')
CRYPTO_SCRIPT = from_hex_str('0x726cb6e0cd8628a1350a611384688911ab75f51b')
LEDGER_SCRIPT = from_hex_str('0xda65b600f7124ce6c79950c1772a36403104f2be')
MANAGEMENT_SCRIPT = from_hex_str('0xfffdc93764dbaddd97c48f252a53ea4643faa3fd')
ORACLE_SCRIPT = from_hex_str('0xfe924b7cfe89ddd271abaf7210a80a7e11178758')
POLICY_SCRIPT = from_hex_str('0xcc5e4edd9f5f8dba8bb65734541df7a1c081c67b')
ROLE_MANAGEMENT = from_hex_str('0x49cf4e5378ffcd4dec034fd98a174c5491e395e2')
STD_LIB_SCRIPT = from_hex_str('0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0')



from typing import Optional

from boa3.neo.vm.VMCode import VMCode
from boa3.neo.vm.type.Integer import Integer


class TryCode(VMCode):
    """
    Represents a Neo VM function try code

    :ivar info: the opcode information of the code
    :ivar data: the data in bytes of the code. Empty byte array by default.
    """

    def __init__(self, except_start_code: Optional[VMCode] = None, finally_start_code: Optional[VMCode] = None):
        """
        :param except_start_code: the first code of the except body
        :type except_start_code: VMCode or None
        """
        self._except_start_code: Optional[VMCode] = except_start_code
        self._finally_start_code: Optional[VMCode] = finally_start_code
        from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
        super().__init__(OpcodeInfo.TRY)

    @property
    def target(self) -> VMCode:
        return self._except_start_code if self._finally_start_code is None else self._finally_start_code

    def set_except_code(self, except_code: VMCode):
        if self._except_start_code is None and except_code is not None:
            self._except_start_code = except_code

    def set_finally_code(self, finally_code: VMCode):
        if self._finally_start_code is None and finally_code is not None:
            self._finally_start_code = finally_code

    @property
    def data(self) -> bytes:
        """
        Gets the Neo VM data of the code

        :return: the formatted data in bytes of the code.
        """
        catch_data: bytes = self._get_raw_data(self._except_start_code)
        finally_data: bytes = self._get_raw_data(self._finally_start_code)

        min_data_len = self._info.data_len // 2
        max_data_len = self._info.max_data_len // 2

        return (self._format_data(catch_data, min_data_len, max_data_len)
                + self._format_data(finally_data, min_data_len, max_data_len))

    def _format_data(self, data: bytes, min_data_len: int, max_data_len: int = -1) -> bytes:
        if max_data_len < min_data_len:
            max_data_len = min_data_len
        mutable_data = bytearray(data)

        if len(mutable_data) < min_data_len:
            mutable_data.extend(bytes(min_data_len))
        elif len(mutable_data) > max_data_len:
            mutable_data = mutable_data[:max_data_len]
        return bytes(mutable_data)

    def _get_raw_data(self, opcode: VMCode) -> bytes:
        """
        Gets the Neo VM raw data of the code

        :return: the unformatted data in bytes of the code.
        """
        min_len = self._info.data_len // 2  # for try opcode, data_len adds catch and finally addresses
        if opcode is None:
            return bytes(min_len)
        else:
            from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
            code_mapping = VMCodeMapping.instance()
            self_start = code_mapping.get_start_address(self)
            target_start = code_mapping.get_start_address(opcode)

            return (Integer(target_start - self_start)
                    .to_byte_array(signed=True, min_length=min_len))

    @property
    def raw_data(self) -> bytes:
        catch_data: bytes = self._get_raw_data(self._except_start_code)
        finally_data: bytes = self._get_raw_data(self._finally_start_code)

        size = max(len(catch_data), len(finally_data))

        return self._format_data(catch_data, size) + self._format_data(finally_data, size)



from typing import List, Tuple

from boa3.model.builtin.method import ScriptHashMethod
from boa3.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.model.type.type import IType
from boa3.neo.vm.opcode.Opcode import Opcode


class ECPointToScriptHashMethod(ScriptHashMethod):

    def __init__(self, data_type: IType = None):
        ec_point = ECPointType.build()
        if not ec_point.is_type_of(data_type):
            data_type = ec_point

        super().__init__(data_type)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3 import constants
        from boa3.model.builtin.interop.crypto.checksigmethod import CheckSigMethod
        from boa3.model.builtin.interop.crypto.sha256method import Sha256Method
        from boa3.model.builtin.interop.crypto.ripemd160method import Ripemd160Method
        from boa3.model.type.type import Type

        # build CheckSig script
        pushdata, pushbytes = Opcode.get_pushdata_and_data_from_size(constants.SIZE_OF_ECPOINT)
        opcodes = [
            Opcode.get_pushdata_and_data(pushdata + pushbytes),
            (Opcode.SWAP, b''),
            Opcode.get_pushdata_and_data(CheckSigMethod.get_raw_bytes()),
            (Opcode.CAT, b''),
            (Opcode.CAT, b''),
        ]

        opcodes.extend([Opcode.get_push_and_data(1),  # pack argument to call syscall
                        (Opcode.PACK, b'')
                        ] + Sha256Method().opcode
                       )
        opcodes.extend([Opcode.get_push_and_data(1),  # pack argument to call syscall
                        (Opcode.PACK, b'')
                        ] + Ripemd160Method().opcode
                       )

        opcodes.extend([
            # limit result to UInt160 length
            Opcode.get_push_and_data(0),
            Opcode.get_push_and_data(constants.SIZE_OF_INT160),
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, Type.bytes.stack_item)
        ])
        return opcodes



from boa3.model.builtin.builtin import Builtin
from boa3.model.standards.neostandard import INeoStandard
from boa3.model.standards.standardmethod import StandardMethod
from boa3.model.type.collection.sequence.uint160type import UInt160Type
from boa3.model.type.type import Type


class Nep17Standard(INeoStandard):
    def __init__(self):
        type_uint160 = UInt160Type.build()

        methods = [
            StandardMethod('symbol', safe=True,
                           args={},
                           return_type=Type.str),
            StandardMethod('decimals', safe=True,
                           args={},
                           return_type=Type.int),
            StandardMethod('totalSupply', safe=True,
                           args={},
                           return_type=Type.int),
            StandardMethod('balanceOf', safe=True,
                           args={'account': type_uint160},
                           return_type=Type.int),
            StandardMethod('transfer',
                           args={'from': type_uint160,
                                 'to': type_uint160,
                                 'amount': Type.int,
                                 'data': Type.any},
                           return_type=Type.bool)
        ]
        events = [
            Builtin.Nep17Transfer
        ]

        super().__init__(methods, events)



import ast
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from boa3 import constants
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.analyser.model.optimizer import ScopeValue, Undefined
from boa3.analyser.model.optimizer.Operation import Operation
from boa3.exception import CompilerWarning
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.method import Method
from boa3.model.module import Module
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.property import Property
from boa3.model.symbol import ISymbol
from boa3.model.type.classes.userclass import UserClass
from boa3.model.type.primitive.primitivetype import PrimitiveType


class AstOptimizer(IAstAnalyser, ast.NodeTransformer):
    """
    This class is responsible for reducing the generated ast.

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar modules: a list with the analysed modules. Empty by default.
    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, analyser, log: bool = False):
        super().__init__(analyser.ast_tree, filename=analyser.filename, log=log)
        self.modules: Dict[str, Module] = {}
        self.symbols: Dict[str, ISymbol] = analyser.symbol_table

        self._is_optimizing: bool = False
        self.has_changes: bool = False
        self.current_scope: ScopeValue = ScopeValue()

        self._current_class: UserClass = None

        self.visit(self._tree)

    @property
    def tree(self) -> ast.AST:
        """
        Gets the analysed abstract syntax tree

        :return: the analysed ast
        """
        return self._tree

    def literal_eval(self, node: ast.AST) -> Any:
        """
        Evaluates an expression node containing a Python expression.

        :param node: the node that will be evaluated
        :return: the evaluated expression if the node is valid. Otherwise, returns Undefined.
        """
        try:
            return ast.literal_eval(node)
        except BaseException:
            return Undefined

    def parse_to_node(self, expression: str, origin: ast.AST = None, is_origin_str: bool = False) -> Union[ast.AST, Sequence[ast.AST]]:
        """
        Parses an expression to an ast.

        :param expression: string expression to be parsed
        :param origin: an existing ast. If not None, the parsed node will have the same location of origin.
        :return: the parsed node
        :rtype: ast.AST or Sequence[ast.AST]
        """
        if is_origin_str:
            expression = "'{0}'".format(expression)

        new_node = self.visit(super().parse_to_node(expression, origin))
        if hasattr(new_node, 'op'):
            new_node.op = Operator.get_operation(new_node.op)

        return new_node

    def reset_state(self):
        self.current_scope.reset()

    def get_symbol_id(self, node: ast.AST) -> Optional[str]:
        parts = []
        cur_node = node
        while isinstance(cur_node, ast.Attribute):
            parts.insert(0, cur_node.attr)
            cur_node = cur_node.value

        if isinstance(cur_node, ast.Name):
            parts.insert(0, cur_node.id)

        return constants.ATTRIBUTE_NAME_SEPARATOR.join(parts)

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:
        if node.name in self.symbols:
            class_symbol = self.symbols[node.name]
            if isinstance(class_symbol, UserClass):
                self._current_class = class_symbol

        self.generic_visit(node)
        self._current_class = None
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        symbols = self.symbols if self._current_class is None else self._current_class.symbols
        method = symbols[node.name]

        if isinstance(method, Property):
            method = method.getter

        if isinstance(method, Method):
            self._is_optimizing = True
            self.has_changes = True

            while self.has_changes:
                self.reset_state()
                self.has_changes = False

                super().generic_visit(node)

        self.end_function_optimization()
        return node

    def end_function_optimization(self):
        self.reset_state()
        self.has_changes = False
        self._is_optimizing = False

    def visit_Assign(self, node: ast.Assign) -> ast.AST:
        super().generic_visit(node)
        self.set_variables_value(node.targets, node.value)
        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AST:
        super().generic_visit(node)
        self.set_variables_value([node.target], node.value)
        return node

    def visit_AugAssign(self, node: ast.AugAssign) -> ast.AST:
        super().generic_visit(node)

        value = self.parse_to_node("x+y", node)
        value.left = node.target
        value.op = node.op
        value.right = node.value

        self.set_variables_value([node.target], value)
        return node

    def set_variables_value(self, targets: List[ast.AST], value: ast.AST):
        new_value = self.literal_eval(value)
        for target in targets:
            if isinstance(target, ast.Name) and isinstance(target.ctx, ast.Store):
                self.current_scope.assign(target.id)

                if new_value is not Undefined:
                    self.current_scope[target.id] = new_value
                elif target.id in self.current_scope:
                    self.current_scope.remove(target.id)

    def visit_BinOp(self, bin_op: ast.BinOp) -> ast.AST:
        """
        Visitor of a binary operation node

        :param bin_op: the python ast binary operation node
        """
        try:
            super().generic_visit(bin_op)

            left_value = self.literal_eval(bin_op.left)
            right_value = self.literal_eval(bin_op.right)

            if (left_value is Undefined and isinstance(bin_op.left, ast.BinOp)
                    and self.is_symmetric_operation(bin_op.op, bin_op.left.op)):
                left_value, right_value = self.reorder_operations(bin_op, bin_op.left)
            elif (right_value is Undefined and isinstance(bin_op.right, ast.BinOp)
                  and self.is_symmetric_operation(bin_op.op, bin_op.right.op)):
                left_value, right_value = self.reorder_operations(bin_op, bin_op.right)

            value = self._evaluate_binary_operation(left_value, right_value, bin_op.op)
            if value is not None:
                self.has_changes = True
                return self.parse_to_node(str(value), bin_op, isinstance(value, str))
            return bin_op
        except ValueError:
            return bin_op

    def is_symmetric_operation(self, first_op: BinaryOperation, second_op: BinaryOperation) -> bool:
        if not isinstance(first_op, BinaryOperation) or not isinstance(second_op, BinaryOperation):
            return False

        operation = type(first_op)
        second_operation = type(second_op)

        if operation != second_operation:
            return False

        return first_op.is_symmetric

    def reorder_operations(self, outer_bin_op: ast.BinOp, inner_bin_op: ast.BinOp) -> Tuple[Any, Any]:
        inner_first_value = self.literal_eval(inner_bin_op.left)
        inner_second_value = self.literal_eval(inner_bin_op.right)

        if (not (isinstance(outer_bin_op.op, BinaryOperation) and outer_bin_op.op.is_symmetric)
                or not (isinstance(outer_bin_op.op, BinaryOperation) and outer_bin_op.op.is_symmetric)):
            return inner_first_value, inner_second_value

        is_left_operand: bool = inner_bin_op is outer_bin_op.left
        other_value = self.literal_eval(outer_bin_op.right if is_left_operand else outer_bin_op.left)

        if inner_first_value is not Undefined or inner_second_value is not Undefined:
            if inner_first_value is Undefined:
                if other_value is Undefined:
                    if is_left_operand:
                        # (x + 1) + y -> (x + y) + 1
                        inner_bin_op.right, outer_bin_op.right = outer_bin_op.right, inner_bin_op.right
                    else:
                        # y + (x + 1) -> 1 + (x + y)
                        inner_bin_op.right, outer_bin_op.left = outer_bin_op.left, inner_bin_op.right
                else:
                    if is_left_operand:
                        # (x + 2) + 1 -> (1 + 2) + x
                        inner_bin_op.left, outer_bin_op.right = outer_bin_op.right, inner_bin_op.left
                    else:
                        # 1 + (x + 2) -> x + (1 + 2)
                        inner_bin_op.left, outer_bin_op.left = outer_bin_op.left, inner_bin_op.left
            else:
                if other_value is Undefined:
                    if is_left_operand:
                        # (1 + x) + y -> (y + x) + 1
                        inner_bin_op.left, outer_bin_op.right = outer_bin_op.right, inner_bin_op.left
                    else:
                        # y + (1 + x) ->  1 + (y + x)
                        inner_bin_op.left, outer_bin_op.left = outer_bin_op.left, inner_bin_op.left
                else:
                    if is_left_operand:
                        # (2 + x) + 1 -> (2 + 1) + x
                        inner_bin_op.right, outer_bin_op.right = outer_bin_op.right, inner_bin_op.right
                    else:
                        # 1 + (2 + x) -> x + (2 + 1)
                        inner_bin_op.right, outer_bin_op.left = outer_bin_op.left, inner_bin_op.right

        super().generic_visit(outer_bin_op)

        return self.literal_eval(outer_bin_op), self.literal_eval(outer_bin_op)

    def _evaluate_binary_operation(self, left: Any, right: Any,
                                   op: Union[ast.operator, BinaryOperation]) -> Optional[Any]:
        operator = Operation.get_operation(op)
        try:
            if operator is Operation.Add:
                return left + right
            if operator is Operation.Sub:
                return left - right
            if operator is Operation.Mult:
                return left * right
            if operator is Operation.FloorDiv:
                return left // right
            if operator is Operation.Mod:
                return left % right
            return None
        except BaseException:
            return None

    def visit_UnaryOp(self, un_op: ast.UnaryOp) -> ast.AST:
        """
        Visitor of a binary operation node

        :param un_op: the python ast binary operation node
        """
        try:
            self.visit(un_op.operand)

            operand_value = ast.literal_eval(un_op.operand)

            value = self._evaluate_unary_operation(operand_value, un_op.op)
            if value is not None:
                self.has_changes = True
                if hasattr(un_op.operand, 'n'):
                    un_op.operand.n = value
                    self.update_line_and_col(un_op.operand, un_op)
                    return un_op.operand
                return self.parse_to_node(str(value), un_op, isinstance(value, str))
            return un_op
        except ValueError:
            return un_op

    def _evaluate_unary_operation(self, operand: Any, op: Union[ast.operator, UnaryOperation]) -> Optional[Any]:
        operator = Operation.get_operation(op)
        try:
            if operator is Operation.Add:
                return +operand
            elif operator is Operation.Sub:
                return -operand
            return None
        except BaseException:
            return None

    def visit_If(self, node: ast.If) -> ast.AST:
        self.visit(node.test)

        if_scope: ScopeValue = self.current_scope.new_scope()
        else_scope: ScopeValue = self.current_scope.new_scope()

        self.current_scope = if_scope
        for stmt in node.body:
            self.visit(stmt)

        if len(node.orelse) > 0:
            self.current_scope = else_scope
            for stmt in node.orelse:
                self.visit(stmt)

        self.current_scope = self.current_scope.previous_scope()
        self.current_scope.update_values(if_scope, else_scope)

        return node

    def visit_loop_body(self, node: ast.AST):
        if not hasattr(node, 'body') or not hasattr(node, 'orelse'):
            self.generic_visit(node)

        loop_scope: ScopeValue = self.current_scope.new_scope()
        # TODO: substitute the variables only if they're not reassigned inside the loop
        loop_scope.reset()

        self.current_scope = loop_scope
        for stmt in node.body:
            self.visit(stmt)

        if len(node.orelse) > 0:
            else_scope: ScopeValue = self.current_scope.new_scope()
            self.current_scope = else_scope

            for stmt in node.orelse:
                self.visit(stmt)

            self.current_scope = else_scope.previous_scope()
            loop_scope.update_values(else_scope, is_loop_scope=True)

        outer_scope = self.current_scope.previous_scope()
        outer_scope.update_values(loop_scope)

        return loop_scope

    def visit_For(self, node: ast.For) -> ast.AST:
        self.visit(node.iter)

        for_scope = self.visit_loop_body(node)
        self.current_scope = for_scope.previous_scope()

        return node

    def visit_While(self, node: ast.While) -> ast.AST:
        while_scope = self.visit_loop_body(node)
        self.visit(node.test)

        self.current_scope = while_scope.previous_scope()

        return node

    def visit_Try(self, node: ast.Try) -> ast.AST:
        outer_scope = self.current_scope
        try_scope: ScopeValue = self.current_scope.new_scope()
        except_scopes: List[ScopeValue] = []

        self.current_scope = try_scope
        for stmt in node.body:
            self.visit(stmt)

        if len(node.handlers) > 0:
            for handler in node.handlers:
                except_scope = outer_scope.new_scope()
                self.current_scope = except_scope

                for stmt in handler.body:
                    self.visit(stmt)

                except_scopes.append(except_scope)

        if len(node.orelse) > 0:
            else_scope = outer_scope.new_scope()
            self.current_scope = else_scope

            for stmt in node.orelse:
                self.visit(stmt)

            except_scopes.append(else_scope)

        self.current_scope = self.current_scope.previous_scope()
        self.current_scope.update_values(try_scope, *except_scopes)

        for stmt in node.finalbody:
            self.visit(stmt)

        return node

    def visit_Name(self, node: ast.Name) -> ast.AST:
        if (isinstance(node.ctx, ast.Load)
                and node.id in self.current_scope
                and isinstance(self.get_type(self.current_scope[node.id]), PrimitiveType)):
            # only values from int, bool, str and bytes types are going to replace the variable
            # TODO: check if it's worth to replace other types
            value = self.current_scope[node.id]
            if isinstance(value, str):
                value = "'{0}'".format(value)
            return self.parse_to_node(str(value), node)
        return node

    def visit_Call(self, node: ast.Call) -> ast.AST:
        # check if the call can be evaluated during compile time
        # TODO: right now only UInt160 and UInt256 constructors are evaluated
        literal_args = []
        args_are_literal = True

        for index, arg in enumerate(node.args.copy()):
            updated_arg = self.visit(arg)  # first try to optimize the arguments
            if updated_arg != arg:
                node.args[index] = updated_arg

            if args_are_literal:
                value = self.literal_eval(updated_arg)
                if value is Undefined:
                    # don't break if one argument is not literal to make sure that all arguments were checked
                    # if they can be optimized
                    args_are_literal = False

                literal_args.append(value)

        if args_are_literal:
            # try to get the result
            try:
                func_id = self.get_symbol_id(node.func)
            except BaseException:
                return node
            func = self.get_symbol(func_id)

            if isinstance(func, IBuiltinMethod):
                try:
                    result = func.evaluate_literal(*literal_args)
                    if result is not Undefined:
                        return self.parse_to_node(str(result), node, is_origin_str=isinstance(result, str))
                except BaseException:
                    self._log_warning(CompilerWarning.InvalidArgument(
                        node.lineno, node.col_offset
                    ))

        return node



import ast
from typing import Dict, List, Tuple

from boa3 import constants
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.model.symbol import ISymbol


class InitStatementsVisitor(IAstAnalyser):
    """
    This class is separate the instructions that should be included in the '_deploy" internal method from those
    from the '_initialize' internal method.

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    """

    def __init__(self, symbols: Dict[str, ISymbol]):
        super().__init__(ast.parse(""), log=True)
        self.symbols = symbols.copy()

        self._deploy_instructions: List[ast.AST] = []
        self._init_instructions: List[ast.AST] = []

    @classmethod
    def separate_global_statements(cls,
                                   symbol_table: Dict[str, ISymbol],
                                   statements: List[ast.AST]) -> Tuple[List[ast.AST], List[ast.AST]]:

        visitor = InitStatementsVisitor(symbol_table)

        root_ast = ast.parse("")
        root_ast.body = statements
        visitor.visit(root_ast)

        return visitor._deploy_instructions, visitor._init_instructions

    def get_symbol_id(self, node: ast.AST, parent: ast.AST) -> str:
        symbol_id = self.visit(node)
        # filter to find the imported variables
        if symbol_id not in self.symbols and hasattr(parent, 'origin') and isinstance(parent.origin, ast.AST):
            symbol_id = '{0}{2}{1}'.format(parent.origin.__hash__(), symbol_id, constants.VARIABLE_NAME_SEPARATOR)
        return symbol_id

    def append_symbol(self, symbol: ISymbol, node: ast.AST):
        if self._should_be_on_deploy(symbol):
            if node not in self._deploy_instructions:
                self._deploy_instructions.append(node)
        else:
            if node not in self._init_instructions:
                self._init_instructions.append(node)

    def _should_be_on_deploy(self, symbol: ISymbol) -> bool:
        return hasattr(symbol, 'is_reassigned') and symbol.is_reassigned

    # region AST visitors

    def visit_Assign(self, node: ast.Assign):
        deploy_var_nodes = []
        init_var_nodes = []

        deploy_symbol = None
        init_symbol = None
        for target in node.targets:
            target_id = self.get_symbol_id(target, node)
            symbol = self.get_symbol(target_id)

            if self._should_be_on_deploy(symbol):
                deploy_var_nodes.append(target)
                deploy_symbol = symbol
            else:
                init_var_nodes.append(target)
                init_symbol = symbol

        if len(deploy_var_nodes) == len(node.targets) or len(init_var_nodes) == len(node.targets):
            self.append_symbol(deploy_symbol if deploy_symbol is not None else init_symbol,
                               node)
        else:
            deploy_ast = ast.Assign(targets=deploy_var_nodes,
                                    value=node.value)
            ast.copy_location(deploy_ast, node)

            init_ast = ast.Assign(targets=init_var_nodes,
                                  value=node.value)
            ast.copy_location(init_ast, node)

            if hasattr(node, 'origin'):
                deploy_ast.origin = node.origin
                init_ast.origin = node.origin

            self.append_symbol(deploy_symbol, deploy_ast)
            self.append_symbol(init_symbol, init_ast)

    def visit_AnnAssign(self, node: ast.AnnAssign):
        var_id = self.get_symbol_id(node.target, node)
        symbol = self.get_symbol(var_id)
        self.append_symbol(symbol, node)

    def visit_AugAssign(self, node: ast.AugAssign):
        var_id = self.get_symbol_id(node.target, node)
        symbol = self.get_symbol(var_id)
        self.append_symbol(symbol, node)

    def visit_Name(self, name: ast.Name) -> str:
        """
        Visitor of a name node

        :param name: the python ast name identifier node
        :return: the identifier of the name
        """
        return name.id

    # endregion



from typing import Any

from boa3.builtin.interop.contract import Contract
from boa3.builtin.type import UInt160


class ContractManagement:
    """
    A class used to represent the ContractManagement native contract
    """

    @classmethod
    def get_minimum_deployment_fee(cls) -> int:
        """
        Gets the minimum fee of contract deployment.

        :return: the minimum fee of contract deployment
        """
        pass

    @classmethod
    def get_contract(cls, script_hash: UInt160) -> Contract:
        """
        Gets a contract with a given hash.

        :param script_hash: a smart contract hash
        :type script_hash: UInt160
        :return: a contract
        :rtype: Contract

        :raise Exception: raised if hash length isn't 20 bytes.
        """
        pass

    @classmethod
    def deploy(cls, nef_file: bytes, manifest: bytes, data: Any = None) -> Contract:
        """
        Creates a smart contract given the script and the manifest.

        :param nef_file: the target smart contract's compiled nef
        :type nef_file: bytes
        :param manifest: the manifest.json that describes how the script should behave
        :type manifest: bytes
        :param data: the parameters for the _deploy function
        :type data: Any

        :return: the contract that was created
        :rtype: Contract

        :raise Exception: raised if the nef or the manifest are not a valid smart contract.
        """
        pass

    @classmethod
    def update(cls, nef_file: bytes, manifest: bytes, data: Any = None):
        """
        Updates the executing smart contract given the script and the manifest.

        :param nef_file: the new smart contract's compiled nef
        :type nef_file: bytes
        :param manifest: the new smart contract's manifest
        :type manifest: bytes
        :param data: the parameters for the _deploy function
        :type data: Any

        :raise Exception: raised if the nef and the manifest are not a valid smart contract or the new contract is the
            same as the old one.
        """
        pass

    @classmethod
    def destroy(cls):
        """
        Destroy the executing smart contract.
        """
        pass



from typing import Any

from boa3.builtin.type import ByteString


def base58_encode(key: bytes) -> str:
    """
    Encodes a bytes value using base58.

    :param key: bytes value to be encoded
    :type key: bytes
    :return: the encoded string
    :rtype: str
    """
    pass


def base58_decode(key: str) -> bytes:
    """
    Decodes a string value encoded with base58.

    :param key: string value to be decoded
    :type key: str
    :return: the decoded bytes
    :rtype: bytes
    """
    pass


def base58_check_encode(key: bytes) -> str:
    """
    Converts a bytes value to its equivalent str representation that is encoded with base-58 digits. The encoded str
    contains the checksum of the binary data.

    :param key: bytes value to be encoded
    :type key: bytes
    :return: the encoded string
    :rtype: str
    """
    pass


def base58_check_decode(key: str) -> bytes:
    """
    Converts the specified str, which encodes binary data as base-58 digits, to an equivalent bytes value. The encoded
    str contains the checksum of the binary data.

    :param key: string value to be decoded
    :type key: str
    :return: the decoded bytes
    :rtype: bytes
    """
    pass


def base64_encode(key: bytes) -> str:
    """
    Encodes a bytes value using base64.

    :param key: bytes value to be encoded
    :type key: bytes
    :return: the encoded string
    :rtype: str
    """
    pass


def base64_decode(key: str) -> bytes:
    """
    Decodes a string value encoded with base64.

    :param key: string value to be decoded
    :type key: str
    :return: the decoded bytes
    :rtype: bytes
    """
    pass


def serialize(item: Any) -> bytes:
    """
    Serializes the given value into its bytes representation.

    :param item: value to be serialized
    :type item: Any
    :return: the serialized value
    :rtype: bytes

    :raise Exception: raised if the item's type is not serializable.
    """
    pass


def deserialize(data: bytes) -> Any:
    """
    Deserializes the given bytes value.

    :param data: serialized value
    :type data: bytes
    :return: the deserialized result
    :rtype: Any

    :raise Exception: raised when the date doesn't represent a serialized value.
    """
    pass


def atoi(value: str, base: int = 10) -> int:
    """
    Converts a character string to a specific base value, decimal or hexadecimal. The default is decimal.

    :param value: the int value as a string
    :type value: str
    :param base: the value base
    :type base: int
    :return: the equivalent value
    :rtype: int

    :raise Exception: raised when base isn't 10 or 16.
    """
    pass


def itoa(value: int, base: int = 10) -> str:
    """
    Converts the specific type of value to a decimal or hexadecimal string. The default is decimal.

    :param value: the int value
    :type value: int
    :param base: the value's base
    :type base: int
    :return: the converted string
    :rtype: int
    """
    pass


def memory_search(mem: ByteString, value: ByteString, start: int = 0, backward: bool = False) -> int:
    """
    Searches for a given value in a given memory.

    :param mem: the memory
    :type mem: bytes or str
    :param value: the value
    :type value: bytes or str
    :param start: the index the search should start from
    :type start: int
    :param backward: whether it should invert the memory
    :type backward: bool

    :return: the index of the value in the memory. Returns -1 if it's not found
    :rtype: int
    """
    pass


def memory_compare(mem1: ByteString, mem2: ByteString) -> int:
    """
    Compares a memory with another one.

    :param mem1: a memory to be compared to another one
    :type mem1: bytes or str
    :param mem2: a memory that will be compared with another one
    :type mem2: bytes or str

    :return: -1 if mem1 precedes mem2, 0 if mem1 and mem2 are equal, 1 if mem1 follows mem2
    :rtype: int
    """
    pass



from typing import Dict

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.nativecontract import LedgerMethod
from boa3.model.variable import Variable


class CurrentHashMethod(LedgerMethod):
    def __init__(self):
        from boa3.model.type.collection.sequence.uint256type import UInt256Type
        identifier = '-get_current_hash'
        syscall = 'currentHash'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=UInt256Type.build())


class CurrentHashProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'current_hash'
        getter = CurrentHashMethod()
        super().__init__(identifier, getter)



from typing import Any, List, Tuple

from boa3.model.type.classes.pythonclass import PythonClass
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.AbiType import AbiType


class NoneType(PythonClass):
    """
    A class used to represent Python None value
    """

    def __init__(self):
        identifier = 'none'
        super().__init__(identifier)

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Void

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            from boa3.model.type.type import Type
            return Type.none

    @classmethod
    def _is_type_of(cls, value: Any):
        from boa3.model.type.type import Type
        return value is None or value is Type.none

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.ISNULL, b'')]


noneType: IType = NoneType()



from typing import Any, Optional

from boa3.model.type.itype import IType
from boa3.neo.vm.type.AbiType import AbiType


class MetaType(IType):
    """
    A class used to represent internal Boa and Python types
    """

    def __init__(self, type_of: IType = None):
        identifier = 'type'
        super().__init__(identifier)
        self._internal_type: Optional[IType] = type_of

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Any

    @classmethod
    def build(cls, value: Any) -> IType:
        if isinstance(value, MetaType):
            return value
        if isinstance(value, IType):
            return cls(value)
        else:
            return metaType

    @classmethod
    def _is_type_of(cls, value: Any) -> bool:
        return isinstance(value, MetaType)

    def is_type_of(self, value: Any) -> bool:
        return isinstance(value, MetaType) and self._internal_type.is_type_of(value._internal_type)

    @property
    def meta_identifier(self) -> str:
        if self.has_meta_type:
            return f'{self._identifier}[{self._internal_type}]'
        else:
            return self._identifier

    @property
    def has_meta_type(self) -> bool:
        return self._internal_type is not None

    @property
    def meta_type(self) -> Optional[IType]:
        return self._internal_type

    @property
    def meta_id(self) -> Optional[None]:
        return self._internal_type.identifier if self._internal_type is not None else None


metaType: IType = MetaType()



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.interop.runtime.triggertype import TriggerType
from boa3.model.variable import Variable


class GetTriggerMethod(InteropMethod):

    def __init__(self, trigger_type: TriggerType):
        identifier = 'get_trigger'
        syscall = 'System.Runtime.GetTrigger'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=trigger_type)



from typing import Any

from boa3.builtin.type import ByteString


class StdLib:
    """
    A class used to represent StdLib native contract
    """

    @classmethod
    def serialize(cls, item: Any) -> bytes:
        """
        Serializes the given value into its bytes representation.

        :param item: value to be serialized
        :type item: Any
        :return: the serialized value
        :rtype: bytes

        :raise Exception: raised if the item's type is not serializable.
        """
        pass

    @classmethod
    def deserialize(cls, data: bytes) -> Any:
        """
        Deserializes the given bytes value.

        :param data: serialized value
        :type data: bytes
        :return: the deserialized result
        :rtype: Any

        :raise Exception: raised when the date doesn't represent a serialized value.
        """
        pass

    @classmethod
    def json_serialize(cls, item: Any) -> str:
        """
        Serializes an item into a json.

        :param item: The item that will be serialized
        :type item: Any
        :return: The serialized item
        :rtype: str

        :raise Exception: raised if the item is an integer value out of the Neo's accepted range, is a dictionary with a
            bytearray key, or isn't serializable.
        """
        pass

    @classmethod
    def json_deserialize(cls, json: str) -> Any:
        """
        Deserializes a json into some valid type.

        :param json: A json that will be deserialized
        :type json: str
        :return: The deserialized json
        :rtype: Any

        :raise Exception: raised if jsons deserialization is not valid.
        """
        pass

    @classmethod
    def base64_decode(cls, key: str) -> bytes:
        """
        Decodes a string value encoded with base64.

        :param key: string value to be decoded
        :type key: str
        :return: the decoded string
        :rtype: bytes
        """
        pass

    @classmethod
    def base64_encode(cls, key: bytes) -> str:
        """
        Encodes a bytes value using base64.

        :param key: bytes value to be encoded
        :type key: bytes
        :return: the encoded string
        :rtype: str
        """
        pass

    @classmethod
    def base58_decode(cls, key: str) -> bytes:
        """
        Decodes a string value encoded with base58.

        :param key: string value to be decoded
        :type key: str
        :return: the decoded bytes
        :rtype: bytes
        """
        pass

    @classmethod
    def base58_encode(cls, key: bytes) -> str:
        """
        Encodes a bytes value using base58.

        :param key: bytes value to be encoded
        :type key: bytes
        :return: the encoded string
        :rtype: str
        """
        pass

    @classmethod
    def base58_check_decode(cls, key: str) -> bytes:
        """
        Converts the specified str, which encodes binary data as base-58 digits, to an equivalent bytes value. The encoded
        str contains the checksum of the binary data.

        :param key: string value to be decoded
        :type key: str
        :return: the decoded bytes
        :rtype: bytes
        """
        pass

    @classmethod
    def base58_check_encode(cls, key: bytes) -> str:
        """
        Converts a bytes value to its equivalent str representation that is encoded with base-58 digits. The encoded str
        contains the checksum of the binary data.

        :param key: bytes value to be encoded
        :type key: bytes
        :return: the encoded string
        :rtype: str
        """
        pass

    @classmethod
    def itoa(cls, value: int, base: int = 10) -> str:
        """
        Converts the specific type of value to a decimal or hexadecimal string. The default is decimal.

        :param value: the int value
        :type value: int
        :param base: the value's base
        :type base: int
        :return: the converted string
        :rtype: int
        """
        pass

    @classmethod
    def atoi(cls, value: str, base: int = 10) -> int:
        """
        Converts a character string to a specific base value, decimal or hexadecimal. The default is decimal.

        :param value: the int value as a string
        :type value: str
        :param base: the value base
        :type base: int
        :return: the equivalent value
        :rtype: int

        :raise Exception: raised when base isn't 10 or 16.
        """
        pass

    @classmethod
    def memory_compare(cls, mem1: ByteString, mem2: ByteString) -> int:
        """
        Compares a memory with another one.

        :param mem1: a memory to be compared to another one
        :type mem1: bytes or str
        :param mem2: a memory that will be compared with another one
        :type mem2: bytes or str

        :return: -1 if mem1 precedes mem2, 0 if mem1 and mem2 are equal, 1 if mem1 follows mem2
        :rtype: int
        """
        pass

    @classmethod
    def memory_search(cls, mem: ByteString, value: ByteString, start: int = 0, backward: bool = False) -> int:
        """
        Searches for a given value in a given memory.

        :param mem: the memory
        :type mem: bytes or str
        :param value: the value
        :type value: bytes or str
        :param start: the index the search should start from
        :type start: int
        :param backward: whether it should invert the memory
        :type backward: bool

        :return: the index of the value in the memory. Returns -1 if it's not found
        :rtype: int
        """
        pass



import ast
from typing import Dict

from boa3.model.builtin.interop.nativecontract import ContractManagementMethod
from boa3.model.variable import Variable


class UpdateMethod(ContractManagementMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'update_contract'
        syscall = 'update'
        args: Dict[str, Variable] = {
            'nef_file': Variable(Type.bytes),
            'manifest': Variable(Type.bytes),
            'data': Variable(Type.any)
        }
        data_default = ast.parse("{0}".format(Type.any.default_value)
                                 ).body[0].value
        super().__init__(identifier, syscall, args, defaults=[data_default], return_type=Type.none)
